/* criteria.c--criteria implementations
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define CRITERIA_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "../util/str.h"
#include "criteria.h"
#include "comparator.h"
#include "../daemon/daemon.h"
#include "../daemon/daemon_err.h"
#include "acap_err.h"
#include "attr.h"

/* We could use OO methods for this.  We could also voluntarily live
 * in New Jersey.  Really; the set of criteria isn't likely to change,
 * and if it does, it's just a couple switch statements to alter; we
 * don't *need* to be obsessively OO about *everything*...
 * (I know!  Plugin criteria!  Not.) */

struct criterion_s {
  criterion_tag_t tag;
  union {
    struct {
      criterion_t    *first;
      criterion_t    *second;
    }               conjunction;
    struct {
      str_t          *attribute;
      union {
	compare_equality_proc_t *equality;
	compare_ordering_proc_t *ordering;
	compare_prefix_proc_t *prefix;
	compare_substring_proc_t *substring;
      }               comparator;
      str_t          *comparator_str;
      str_t          *value;
      void           *cache;
    }               comparison;
    struct {
      unsigned        low,
                      high;
      str_t          *modtime;
    }               range;
  }               value;
};

static          result_t
criterion_alloc(criterion_tag_t tag,
		criterion_t ** dest)
{
  I(dest);

  *dest = malloc(sizeof(criterion_t));
  if (!*dest)
    return errno;
  memset(*dest, 0L, sizeof(criterion_t));
  (*dest)->tag = tag;
  return OK;
}

result_t
_criterion_conjunction_alloc(criterion_tag_t tag,
			     criterion_t * first,
			     criterion_t * second,
			     criterion_t ** dest)
{
  result_t        result;

  I(dest);

  result = criterion_alloc(tag, dest);
  if (result)
    return result;
  (*dest)->value.conjunction.first = first;
  (*dest)->value.conjunction.second = second;
  return OK;
}

result_t
_criterion_comparison_alloc(criterion_tag_t tag,
			    str_t * attribute,
			    str_t * comparator,
			    str_t * value,
			    criterion_t ** dest)
{
  result_t        result;

  I(attribute);
  I(comparator);
  I(dest);

  result = criterion_alloc(tag, dest);
  if (result)
    return result;

  switch (tag) {
  case CRITERION_COMPARE:
  case CRITERION_COMPARESTRICT:
    result = comparator_lookup_ordering
      (comparator, &(*dest)->value.comparison.comparator.ordering);
    break;
  case CRITERION_EQUAL:
    result = comparator_lookup_equality
      (comparator, &(*dest)->value.comparison.comparator.equality);
    break;
  case CRITERION_PREFIX:
    result = comparator_lookup_prefix
      (comparator, &(*dest)->value.comparison.comparator.prefix);
    break;
  case CRITERION_SUBSTRING:
    result = comparator_lookup_substring
      (comparator, &(*dest)->value.comparison.comparator.substring);
    break;
  default:
    daemon_fatal(DAEMON_ERR_INTERNAL, "Bad tag for comparison criterion");
  }

  if (result) {
    free(*dest);
    *dest = NULL;
    return result;
  }
  (*dest)->value.comparison.attribute = attribute;
  (*dest)->value.comparison.comparator_str = comparator;
  (*dest)->value.comparison.value = value;
  (*dest)->value.comparison.cache = NULL;

  return OK;
}

result_t
criterion_range_alloc(unsigned low, unsigned high,
		      str_t * modtime, criterion_t ** dest)
{
  result_t        result;

  I(modtime);
  I(dest);

  result = criterion_alloc(CRITERION_RANGE, dest);
  if (result)
    return result;

  (*dest)->value.range.low = low;
  (*dest)->value.range.high = high;
  (*dest)->value.range.modtime = modtime;

  return OK;
}

result_t
criterion_free(criterion_t ** criterion)
{
  criterion_t    *crit;

  criterion_t    *todo = NULL;

  /*
   * Okay.  This is... a little ugly.
   * 
   * The general idea is that we walk the tree, depth-first, inserting branches
   * into the todo list.  When we free a leaf, we pop the todo list, free the
   * node, and continue down the new branch.
   * 
   * We do this without allocing by linking via the "second" field of
   * conjunctions.
   * 
   * So, if we see an AND or an OR, we set its second field to the current TODO
   * list and change it into a NOT.  If we see anything else, we free it; if
   * it's a NOT, we use its "first" field as our next piece of data, and if
   * it's anything else, we pop the todo stack (which should be a NOT); the
   * new todo stack is the NOT's "second" field.
   * 
   * Clear?
   * 
   * Oh--we do this rigamarole so that we can free all the criteria without
   * resorting to recursion; normally, I'd be first in line advocating the
   * clean recursive solution, but not in a threaded environment.  So,
   * instead, we have this iterative mess.
   */

  I(criterion);
  I(*criterion);
  crit = *criterion;

  while (crit) {
    switch (crit->tag) {
    case CRITERION_AND:
    case CRITERION_OR:
      I(crit->value.conjunction.first);
      I(crit->value.conjunction.second);
      {
	criterion_t    *next = crit->value.conjunction.second;

	crit->value.conjunction.second = todo;
	crit->tag = CRITERION_NOT;
	todo = crit;
	crit = next;
      }
      break;
    case CRITERION_NOT:
      I(crit->value.conjunction.first);
      I(!crit->value.conjunction.second);
      {
	criterion_t    *next = crit->value.conjunction.first;

	free(crit);
	crit = next;
      }
      break;
    case CRITERION_ALL:
      I(!crit->value.conjunction.first);
      I(!crit->value.conjunction.second);
      free(crit);
      crit = todo;
      if (crit) {
	todo = crit->value.conjunction.second;
	crit->value.conjunction.second = NULL;
      }
      break;
    case CRITERION_COMPARE:
    case CRITERION_COMPARESTRICT:
    case CRITERION_EQUAL:
    case CRITERION_PREFIX:
    case CRITERION_SUBSTRING:
      I(crit->value.comparison.attribute);
      I(crit->value.comparison.comparator.equality);
      I(crit->value.comparison.comparator_str);
      str_free(&crit->value.comparison.attribute);
      str_free(&crit->value.comparison.comparator_str);
      if (crit->value.comparison.value)
	str_free(&crit->value.comparison.value);
      if (crit->value.comparison.cache)
	free(&crit->value.comparison.cache);
      free(crit);
      crit = todo;
      if (crit) {
	todo = crit->value.conjunction.second;
	crit->value.conjunction.second = NULL;
      }
      break;
    case CRITERION_RANGE:
      I(crit->value.range.modtime);
      str_free(&crit->value.range.modtime);
      free(crit);
      crit = todo;
      if (crit) {
	todo = crit->value.conjunction.second;
	crit->value.conjunction.second = NULL;
      }
      break;
    default:
      daemon_fatal(DAEMON_ERR_INTERNAL, "Bad criterion tag");
    }
  }

  *criterion = NULL;
  return OK;
}

static result_t
check_compare(criterion_t * criterion,
	      str_t * value,
	      bool_t * dest)
{
  result_t        result;

  compare_result_t compare_result;

  I(criterion);
  I(dest);

  result = criterion->value.comparison.comparator.ordering
    (criterion->value.comparison.value, value, &compare_result);

  if (compare_result == COMPARE_LESS
      || compare_result == COMPARE_EQUAL)
    *dest = TRUE;
  else
    *dest = FALSE;

  return result;
}

static result_t
check_comparestrict(criterion_t * criterion,
		    str_t * value,
		    bool_t * dest)
{
  result_t        result;

  compare_result_t compare_result;

  I(criterion);
  I(dest);

  result = criterion->value.comparison.comparator.ordering
    (criterion->value.comparison.value, value, &compare_result);

  if (compare_result == COMPARE_LESS)
    *dest = TRUE;
  else
    *dest = FALSE;

  return result;
}

static result_t
check_equal(criterion_t * criterion,
	    str_t * value,
	    bool_t * dest)
{
  result_t        result;

  eq_result_t     eq_result;

  I(criterion);
  I(dest);

  result = criterion->value.comparison.comparator.equality
    (criterion->value.comparison.value, value, &eq_result);

  if (eq_result == EQ_IS_EQUAL)
    *dest = TRUE;
  else
    *dest = FALSE;

  return result;
}

static result_t
check_prefix(criterion_t * criterion,
	     str_t * value,
	     bool_t * dest)
{
  result_t        result;

  eq_result_t     eq_result;

  I(criterion);
  I(dest);

  result = criterion->value.comparison.comparator.prefix
    (criterion->value.comparison.value, value, &eq_result);

  if (eq_result == EQ_IS_EQUAL)
    *dest = TRUE;
  else
    *dest = FALSE;

  return result;
}

static result_t
check_substring(criterion_t * criterion,
		str_t * value,
		bool_t * dest)
{
  result_t        result;

  eq_result_t     eq_result;

  I(criterion);
  I(dest);

  result = criterion->value.comparison.comparator.substring
    (criterion->value.comparison.value,
     value,
     &criterion->value.comparison.cache,
     &eq_result);

  if (eq_result == EQ_IS_EQUAL)
    *dest = TRUE;
  else
    *dest = FALSE;

  return result;
}

result_t
criterion_eval(criterion_t * criterion,	/* current criterion to process */
	       sllist_t ** stack_node_cache,	/* source of unused sllist_t
						 * *s */
	       const vent_t * entry,
	       unsigned index,
	       bool_t * dest)
{
  result_t        result;

  const attr_t   *attr;		/* attribute being checked */

  result_t        (*check_proc)	/* function used for comparisons */
                  (criterion_t * criterion,
		                   str_t * value,
		                   bool_t * dest);

  strlist_iterator_t strlist_iter;	/* used for walking value lists */

  sllist_t       *criterion_stack = NULL;	/* list of criterion we've
						 * looked at should be only
						 * AND, OR, and NOT */

  /*
   * The game plan: criterion points to the criterion we're currently
   * considering.  If it's an AND, OR, or NOT, we push it onto the criterion
   * stack, and process one of the kids.  Anything else, we process it, and
   * then loop on the criterion stack--if it's empty, we're done, if it's a
   * NOT, we invert and keep going, and if it's an AND or an OR we perform
   * short circuit evaluation, and either keep going up or go down the other
   * branch; either way, we remove the conjunction from the stack.
   * 
   * We use stack_node_cache to keep stack nodes around which aren't being used;
   * when it runs low, we just alloc more, and when we remove nodes from the
   * criterion_stack we put them onto the stack_node_cache, to be reused
   * later (by this or other criterion_eval invocations), avoiding lots of
   * mallocing.
   * 
   * We *could* also do this trick by rewriting the criterion tree as we
   * traversed it; it would mean we'd never allocate memory in order to
   * process a criterion and we'd still use constant space on the C code
   * stack, but it's more than a little dirty, and would mean we couldn't
   * ever use multiple processors (to look ahead into the far distant
   * future).  The stack_node_cache should make this fast enough anyway...
   */

  I(criterion);
  I(stack_node_cache);
  I(entry);
  I(dest);

  while (criterion) {
    check_proc = NULL;
    switch (criterion->tag) {
    case CRITERION_AND:
    case CRITERION_OR:
    case CRITERION_NOT:
      {
	/*
	 * Push the criterion onto the criterion_stack, and process the LHS
	 */
	sllist_t       *node;

	if (*stack_node_cache) {
	  node = *stack_node_cache;
	  *stack_node_cache = sllist_NEXT(node);
	} else {
	  node = (sllist_t *) malloc(sizeof(sllist_t));
	  if (!node) {
	    *stack_node_cache = criterion_stack;	/* recycle the nodes */
	    return errno;
	  }
	}
	sllist_SET_NEXT(node, criterion_stack);
	sllist_SET_DATA(node, criterion);
	criterion_stack = node;
	criterion = criterion->value.conjunction.first;
	I(criterion);
	continue;		/* while(criterion) */
      }
    case CRITERION_ALL:
      *dest = TRUE;
      break;
    case CRITERION_COMPARE:
      check_proc = &check_compare;
      /* fallthrough */
    case CRITERION_COMPARESTRICT:
      if (check_proc == NULL)
	check_proc = &check_comparestrict;
      /* fallthrough */
    case CRITERION_EQUAL:
      if (check_proc == NULL)
	check_proc = &check_equal;
      /* fallthrough */
    case CRITERION_PREFIX:
      if (check_proc == NULL)
	check_proc = &check_prefix;
      /* fallthrough */
    case CRITERION_SUBSTRING:
      if (check_proc == NULL)
	check_proc = &check_substring;

      /* Okay, we get here if it's a comparison criterion of some sort. */

      result = vent_lookup_attr(entry,
				criterion->value.comparison.attribute,
				&attr);

      I(!attr || attr_STATE(attr) != ATTR_DEFAULT);

      *dest = FALSE;		/* assume no match */
      if (attr && (attr_STATE(attr) == ATTR_SINGLE
		   || attr_STATE(attr) == ATTR_MULTI))
	for (strlist_ITER_BEGIN(attr_VALUE(attr), &strlist_iter);
	     (*dest == FALSE
	      && result == OK
	      && !strlist_ITER_ATEND(&strlist_iter));
	     strlist_ITER_INCR(&strlist_iter)) {
	  result = check_proc(criterion,
			      strlist_ITER_VALUE(&strlist_iter),
			      dest);
	}			/* for(strlist iterator) */
      else			/* ! attr || attr_STATE(attr) != string */
	result = check_proc(criterion, NULL, dest);

      if (result != OK) {
	/* TODO: Cleanup stack */
	return result;
      }
      break;
    case CRITERION_RANGE:
      /* TODO: Modtime check, index==NULL check */
      *dest = (criterion->value.range.low <= index
	       && index <= criterion->value.range.high);
      break;
    }

    /*
     * Okay, at this point we've processed a basic criterion, and the result
     * of that is in *dest.  Now we need to go back to the criterion stack
     * and see just what we're supposed to do next...
     */

    criterion = NULL;		/* In case we don't get a new criterion... */

    while (criterion_stack) {
      /* First, get the new criterion */
      sllist_t       *node = criterion_stack;

      criterion = sllist_DATA(node);
      criterion_stack = sllist_NEXT(node);
      sllist_SET_NEXT(node, *stack_node_cache);
      *stack_node_cache = node;

      /* Now, process it. */
      I(criterion);
      I(criterion->tag == CRITERION_AND
	|| criterion->tag == CRITERION_OR
	|| criterion->tag == CRITERION_NOT);
      if (criterion->tag == CRITERION_NOT) {
	*dest = !*dest;
	/* Simple enough; fall through to criterion stack walk. */
      } else if ((criterion->tag == CRITERION_AND && *dest == TRUE)
		 || (criterion->tag == CRITERION_OR && *dest == FALSE)) {
	/*
	 * In these cases, we can't do short-circuit processing; our result
	 * is the RHS of the expression.
	 */
	criterion = criterion->value.conjunction.second;
	I(criterion);
	break;			/* break the criterion_stack walk; drop into
				 * criterion processing. */
      }
      /*
       * If we get here, then either the criterion was a NOT, or it was an
       * AND or an OR which was short-circuited. Either way, we just keep
       * going up the criterion_stack.  We set criterion to NULL to ensure
       * that we don't use it if the criterion_stack is empty.
       */
      criterion = NULL;
    }				/* while(criterion_stack) */
  }				/* while(criterion) */

  I(!criterion_stack);

  /* Whew! */
  return OK;
}
