/* dataset.h--dataset interface
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef DATASET_H
#define DATASET_H

/* A dataset_t is a moveable structure containing all the information
 * about a dataset: what it inherits from, what its base ACL is, its
 * attribute ACLs, things like that.  Nothing terribly earth-shaking,
 * but it's necessary... */

#include "../util/str.h"
#include "strlist.h"
#include "stamp.h"

/* Note: there're parts of this code which're fairly redundant
 * (iterator initialization, for example), especially when the entry.h
 * code is included.  Maybe this should be abstracted a little?
 * Hrm. */

typedef struct {
  hashvalue_t     attr_name_hash;
  uint32_t        attr_name;
  uint32_t        acls;
}               dataset_acl_ref_t;

#define dataset_ACL_NAME_HASH(a) ((a)->attr_name_hash)
#define dataset_ACL_NAME(a) ((str_t *)((caddr_t)(&(a)->attr_name)	\
				       + ntohl((a)->attr_name)))
#define dataset_ACL_ACLS(a) ((strlist_t *)((caddr_t)(&(a)->acls)	\
					   + ntohl((a)->acls)))

typedef struct dataset_s {
  uint32_t        size;		/* amount of memory used by the dataset */
  stamp_t         modtime;	/* time this dataset object was created */
  uint32_t        datapath;	/* path to the dataset's data */
  uint32_t        inherit;	/* path to the dataset this inherits from */
  uint32_t        base_acl;	/* the base acl of this dataset */
  uint32_t        n_acls;	/* the number of attribute acls */
  uint32_t        acls;		/* the attribute acls of the dataset */
}               dataset_t;

#define dataset_SIZE(d) (ntohl((d)->size))
#define dataset_MODTIME(d) (&(d)->modtime)
#define dataset_DATAPATH(d) ((str_t *)((caddr_t)(&(d)->datapath)	\
				       + ntohl((d)->datapath)))
#define dataset_DEFAULT_INHERIT(d) ((signed)ntohl((d)->inherit) == -1)
#define dataset_INHERIT(d) ((str_t *)(((d)->inherit			\
                                       && ((signed)ntohl((d)->inherit)	\
					   != -1))			\
				      ? ((caddr_t)(&(d)->inherit)	\
					 + ntohl((d)->inherit))		\
				      : NULL))
#define dataset_BASE_ACL(d) ((str_t *)((d)->base_acl			\
				       ? ((caddr_t)(&(d)->base_acl)	\
					  + ntohl((d)->base_acl))	\
				       : NULL))
typedef struct {
  dataset_acl_ref_t *loc,
                 *end;
}               dataset_acl_iterator_t;

#define dataset_ACL_REFS(d) ((dataset_acl_ref_t *)	\
			     ((caddr_t)(&(d)->acls)	\
			      + ntohl((d)->acls)))
#define dataset_ACL_ITER_BEGIN(d, i)		\
	((i)->loc = (dataset_ACL_REFS(d)		\
		     + ntohl((d)->n_acls) - 1),	\
	 (i)->end = (dataset_ACL_REFS(d) - 1))
#define dataset_ACL_ITER_ATEND(i) ((i)->loc <= (i)->end)
#define dataset_ACL_ITER_INCR(i) ((i)->loc--)
#define dataset_ACL_ITER_VALUE(i) ((i)->loc)

typedef struct {
  uint32_t        size;		/* amount of memory actually alloced */
}               dataset_create_state_t;

/* Construction */
extern          result_t
dataset_alloc(const stamp_t * modtime,
	      const str_t * datapath,
	      bool_t default_inherit,
	      const str_t * inherit,
	      const str_t * base_acl,
	      dataset_t ** dataset,
	      dataset_create_state_t * create_state);

/* After the basic dataset object's been created,
 * the caller should append ACLs (by calling dataset_add_attr(),
 * followed by calls to dataset_add_attr_acl()) and then entries
 * with dataset_add_entry(), and then call dataset_complete()
 * to finish it off.
 */
extern result_t
dataset_add_attr(const str_t * attribute,
		 dataset_t ** dataset,
		 dataset_create_state_t * create_state);

extern result_t
dataset_add_attr_acl(const str_t * acl,
		     dataset_t ** dataset,
		     dataset_create_state_t * create_state);

extern result_t
dataset_complete(dataset_t ** dataset,
		 dataset_create_state_t * create_state);

/* This had *better* only be called for datasets which're on the heap... */
extern result_t
                dataset_free(dataset_t ** dataset);

extern result_t
dataset_rewrite(const dataset_t * original,
		const stamp_t * stamp,
		const str_t * datapath,
		bool_t default_inherit,
		const str_t * inherit,
		const str_t * base_acl,
		dataset_t ** dest);

#endif				/* DATASET_H */
