/* dataset_db.h--master dataset db interface
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef DATASET_DB_H
#define DATASET_DB_H

/* Okay, the idea here is that we've got dataset databases; that is,
 * databases which contain information about datasets, and how to
 * get from dataset to dataset, etc.  There will always be at least
 * one of these; I'm keeping the possibility open for multiple
 * dataset databases, though, because it might be very useful at
 * some point to be able to split the hierarchy of datasets up
 * amongst multiple databases.
 *
 * These are reference counted.
 *
 * These are threadsafe--any locking is taken care of internally.
 */

#include "dataset.h"
#include "../util/hash.h"
#include "../util/refcount.h"
#include "trx.h"

struct dataset_ref_s;

struct __db;

typedef struct dataset_db_s {
  refcount_t      refcount;
  hash_entry_t   *hashent;
  struct __db    *db;
  pthread_mutex_t mutex;
  hash_table_t    dataset_refs;
}               dataset_db_t;

#define dataset_db_NAME(d) ((str_t *)hash_GET_KEY((d)->hashent))

#if defined(DATASET_DB_C)
#define dataset_db_MUTEX(d) (&(d)->mutex)
#define dataset_db_DATASET_REFS(d) (&(d)->dataset_refs)
#endif				/* defined(DATASET_DB_C) */

#if defined(DATASET_DB_C) || defined(DATASET_DB_FAC_C)
#define dataset_db_REFCOUNT(d) (&(d)->refcount)
#define dataset_db_HASHENT(d) ((d)->hashent)

/* This returns a new dataset_db, complete with incremented refcount. */
extern          result_t
dataset_db_alloc(hash_entry_t * hashent,
		 dataset_db_t ** dest);

extern result_t dataset_db_free(dataset_db_t ** db);

#endif				/* defined(DATASET_DB_C) ||
				 * defined(DATASET_DB_FAC_C) */

extern result_t dataset_db_ref(dataset_db_t * db);

extern result_t dataset_db_unref(dataset_db_t ** db);

/* This looks up a dataset key in a dataset db.
 * The data is in its own little buffer; it should be free()d
 * with dataset_free() when it's no longer in use. */
extern result_t
dataset_db_lookup(dataset_db_t * db,
		  trx_t * trx,
		  dataset_key_t dataset_key,
		  struct dataset_ref_s ** dataset);

#if defined(DATASET_DB_C) || defined(TRAVERSE_C) || defined(DATASET_REF_C)
/* These are only useable by the traversal and dataset_ref code. */
extern result_t
dataset_db_store(dataset_db_t * db,
		 trx_t * trx,
		 dataset_key_t * dataset_key,
		 dataset_t * dataset);

extern result_t
dataset_db_replace(dataset_db_t * db,
		   trx_t * trx,
		   dataset_key_t dataset_key,
		   dataset_t * dataset);

extern result_t
dataset_db_remove(dataset_db_t * db,
		  trx_t * trx,
		  dataset_key_t dataset_key);

extern result_t
dataset_db_lookup_entry(dataset_db_t * db,
			trx_t * trx,
			dataset_key_t dataset_key,
			const str_t * name,
			dataset_key_t * entry_key);

extern result_t
dataset_db_replace_entry(dataset_db_t * db,
			 trx_t * trx,
			 dataset_key_t dataset_key,
			 const str_t * name,
			 dataset_key_t entry_key);

extern result_t
dataset_db_remove_entry(dataset_db_t * db,
			trx_t * trx,
			dataset_key_t dataset_key,
			const str_t * name);

#endif				/* defined(DATASET_DB_C) ||

				 * 
				 * 
				 * 
				 * defined(TRAVERSE_C) || defined(DATASET_REF_C) */

#if defined(DATASET_DB_C) || defined(DATASET_REF_C)
extern result_t
dataset_db_check_norefs(dataset_db_t * db,
			struct dataset_ref_s * dataset_ref);

#endif				/* defined(DATASET_DB_C) ||

				 * 
				 * 
				 * 
				 * defined(DATASET_REF_C) */

extern result_t
                dataset_db_sync(dataset_db_t * db);

#endif				/* DATASET_DB_H */
