/* dataset_db_fac.c--dataset db factory implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define DATASET_DB_FAC_C
#include <config.h>
#include <pthread.h>
#include "dataset_db_fac.h"
#include "dataset_db.h"
#include "../util/hash.h"
#include "../daemon/daemon_err.h"
#include "../daemon/daemon.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_acapd_dataset_db)

static pthread_mutex_t dataset_db_fac_mutex = PTHREAD_MUTEX_INITIALIZER;

static hash_table_t dataset_db_fac_hash = HASH_INITIALIZER(&hash_str_procs);

result_t
dataset_db_fac_lookup(const str_t * name,
		      dataset_db_t ** dest)
{
  result_t        result;

  hash_did_insert_t inserted;

  hash_entry_t   *hash_entry;

  I(name);
  I(dest);

  *dest = NULL;

  pthread_mutex_lock(&dataset_db_fac_mutex);

  result = hash_find(&dataset_db_fac_hash,
		     name,
		     HASH_INSERT,
		     &inserted,
		     &hash_entry);
  if (result != OK)
    goto cleanup;

  if (inserted == HASH_DID_INSERT) {
    /*
     * Technically, this ought to not keep the whole map locked while
     * creating particular dataset_db; this should be split into seperate
     * allocation and initialization.  OTOH, I don't expect this to happen
     * all that often... Still, TODO.
     */
    result = dataset_db_alloc(hash_entry, dest);
    /*
     * This will have come back from dataset_db_alloc with an incremented
     * refcount.
     */
    if (result != OK) {
      hash_delete_entry(hash_entry);
      goto cleanup;
    }
    hash_SET_VALUE(hash_entry, *dest);
  } else {
    *dest = hash_GET_VALUE(hash_entry);
    dataset_db_ref(*dest);
  }

cleanup:
  pthread_mutex_unlock(&dataset_db_fac_mutex);

  return result;
}

/* Used by dataset_dbs which think they're out of references... */
result_t
dataset_db_fac_check_norefs(dataset_db_t * db)
{
  I(db);

  pthread_mutex_lock(&dataset_db_fac_mutex);

  refcount_UNREF_ELSE(dataset_db_REFCOUNT(db),
		      (hash_delete_entry(dataset_db_HASHENT(db)),
		       pthread_mutex_unlock(&dataset_db_fac_mutex),
		       dataset_db_free(&db)),
		      pthread_mutex_unlock(&dataset_db_fac_mutex));

  return OK;
}
