/* dataset_ref.h - dataset_ref
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef DATASET_REF_H
#define DATASET_REF_H

/* entry_ref and dataset_ref are two highly coupled types
 * which form the datastructure used for ACAP command
 * interlocking. */

#include <pthread.h>
#include "dataset.h"
#include "../util/rselock.h"
#include "../util/str.h"
#include "../util/hash.h"
#include "../util/refcount.h"
#include "dataset.h"
#include "trx.h"

struct entry_ref_s;

struct dataset_data_s;

struct dataset_db_s;

typedef struct dataset_ref_s {
  refcount_t      refcount;
  struct dataset_db_s *db;
  hash_entry_t   *hashent;	/* entry in the db */

  pthread_mutex_t entries_mutex;
  hash_table_t    entries;	/* str_t => entry_ref_t * */

  pthread_mutex_t dataset_mutex;
  dataset_t      *dataset;
  dataset_key_t   key;

  pthread_mutex_t data_mutex;
  struct dataset_data_s *data;
}               dataset_ref_t;

#if defined(DATASET_DB_C) || defined(DATASET_REF_C)
#define dataset_ref_REFCOUNT(d) (&(d)->refcount)
#define dataset_ref_HASHENT(d) ((d)->hashent)
#endif				/* defined(DATASET_DB_C) ||
				 * defined(DATASET_REF_C) */

#if defined(DATASET_REF_C)
#define dataset_ref_DATA_MUTEX(d) (&(d)->data_mutex)
#define dataset_ref_DATA(d) ((d)->data)

#define dataset_ref_ENTRIES(d) (&(d)->entries)
#define dataset_ref_ENTRIES_MUTEX(d) (&(d)->entries_mutex)
#endif				/* defined(DATASET_REF_C) */

#define dataset_ref_DB(d) ((d)->db)
#define dataset_ref_KEY(d) ((d)->key)

#define dataset_ref_DATASET_MUTEX(d) (&(d)->dataset_mutex)

/* Whether this succeeds or not, it takes ownership of the dataset_t.
 * If it succeeds, it returns a new dataset_ref with a refcount of one. */
result_t
dataset_ref_alloc(struct dataset_db_s * db,
		  hash_entry_t * hashent,
		  dataset_key_t key,
		  dataset_t * dataset,
		  dataset_ref_t ** dest);

result_t        dataset_ref_free(dataset_ref_t ** dataset_ref);

/* This add a reference to a dataset_ref_t. */
extern result_t dataset_ref_ref(dataset_ref_t * dataset);

/* This removes a reference from an
 * dataset_ref_t, setting the entry to NULL */
extern result_t dataset_ref_unref(dataset_ref_t ** dataset);

/* This looks up an entry_ref,
 * returning it with an incremented refcount. */
extern result_t
dataset_ref_lookup_entry(dataset_ref_t * dataset_ref,
			 trx_t * trx,
			 const str_t * name,
			 struct entry_ref_s ** entry_ref);

/* This takes care of renaming an entry in the ref
 * tree, as well as updating the filesystem paths. */
extern result_t
dataset_ref_rename_entry(dataset_ref_t * dataset,
			 str_t * oldname,
			 str_t * newname);

#if defined(DATASET_REF_C) || defined(ENTRY_REF_C)
extern result_t
dataset_ref_check_entry_norefs(dataset_ref_t * dataset_ref,
			       struct entry_ref_s * entry_ref);

#endif				/* defined(DATASET_REF_C) ||

				 * defined(ENTRY_REF_C) */

#if defined(DATASET_REF_C) || defined(DATASET_DATA_C)
extern result_t
dataset_ref_check_data_norefs(dataset_ref_t * dataset,
			      refcount_t * refcount,
			      struct dataset_data_s * dataset_data);

#endif				/* defined(DATASET_REF_C) ||

				 * defined(DATASET_DATA_C) */


/* This gets a dataset_ref's associated data object,
 * with an updated refcount. */
extern result_t
dataset_ref_get_dataset_data(dataset_ref_t * dataset,
			     struct dataset_data_s ** dataset_data);

/* Sets the cache size */
extern result_t dataset_ref_set_cachesize(unsigned newsize);

/* Sets the inheritance path for a dataset; assumes no locks */
extern result_t
dataset_ref_set_inherit(dataset_ref_t * dataset_ref,
			trx_t * trx,
			bool_t non_default,
			str_t * path);

/* Gets the dataset_t of the dataset */
extern result_t
dataset_ref_get_dataset(dataset_ref_t * dataset_ref,
			trx_t * trx,
			dataset_t ** dest);

#endif				/* DATASET_REF_H */
