/* lex.acap.c - acap lexer
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define LEX_ACAP_C
#include <config.h>
#include <ctype.h>
#include <stdlib.h>
#include "../util/str.h"
#include "client.h"
#include "parse.acap.h"
#include "token.acap.h"
#include "../daemon/daemon_err.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (acap_debug)

/* Normally, I'd use lex or flex for something like this, but I really
 * want a reentrant (threadsafe) parser, and that's Hard to do without
 * flex and C++. */

int
acap_lex(YYSTYPE * lvalp, client_t * client)
{

  result_t result = OK;

#define ERR() {								\
		VL(("Error encountered in lexer\n"));			\
		client_SET_STATE(client, CLIENT_RECOVER);		\
		return (result == DAEMON_ERR_EOF ? EOF : ERROR);	\
  	      }

#define ERR_PUSHBACK() {				\
    		conn_PUSHBACK(client_CONN(client), c);	\
		ERR();					\
  	      }

  char            c,

                  buffer[ACAP_MAX_QSTR_LEN],	/* big enough for everything */

                 *buff_p = buffer,

                 *buff_end = buffer + ACAP_MAX_QSTR_LEN;

  unsigned long   count = 0;	/* used for reading numbers */

  bool_t          synchronizing = FALSE;	/* sync literal? */

  I(lvalp);
  I(client);

  VL(("Entering ACAP lexer; client %p, conn %p\n", client, client_CONN(client)));

  while (1) {
    VL(("Getting a char: input.buffer_position = %p, buffer_end = %p\n", client->conn->input.buffer_position, client->conn->input.buffer_end));
    result = conn_getc(client_CONN(client), &c);
    VL(("Lexing client %p; state is %d, char is 0x%02x\n", client, client_STATE(client), c));
    if (result != OK)
      ERR();
    switch (client_STATE(client)) {
    case CLIENT_TAG:
      if (buff_p != buffer && c == ' ') {
	/* We've got the tag. */
	lvalp->str = NULL;
	result = str_alloc(buff_p - buffer, buffer, &lvalp->str);
	if (result != OK)
	  ERR_PUSHBACK();
	if (client_parser_add_cleanup(client, &free, lvalp->str)) {
	  str_free(&lvalp->str);
	  ERR_PUSHBACK();
	}
	VL(("Got a tag: %s\n", str_DATA(lvalp->str)));

	conn_PUSHBACK(client_CONN(client), c);
	client_SET_STATE(client, CLIENT_NORMAL);
	return TAG;
      }
      if (buff_end <= buff_p || (!isalnum((unsigned char) c) && c != '.'))
	ERR_PUSHBACK();

      *buff_p++ = c;
      break;
    case CLIENT_RECOVER:
      if (c == '\r')
	client_SET_STATE(client, CLIENT_RECOVER_CR);
      break;
    case CLIENT_RECOVER_CR:
      if (c == '\n')
	client_SET_STATE(client, CLIENT_TAG);
      return CRLF;
    case CLIENT_CR:
      if (c == '\n') {
	client_SET_STATE(client, CLIENT_TAG);
	return CRLF;
      }
      /* otherwise, life is bad */
      ERR_PUSHBACK();
    case CLIENT_QSTR:
      if (c == '\"') {
	/* End of the string */
	lvalp->str = NULL;
	result = str_alloc(buff_p - buffer, buffer, &lvalp->str);
	if (result != OK)
	  ERR_PUSHBACK();
	if (client_parser_add_cleanup(client, &free, lvalp->str)) {
	  str_free(&lvalp->str);
	  ERR_PUSHBACK();
	}
	VL(("Got a string: %s\n", str_DATA(lvalp->str)));
	client_SET_STATE(client, CLIENT_NORMAL);
	return STRING;
      }
      if (c == '\0'
	  || c == '\r'
	  || c == '\n'

#ifdef __CHAR_UNSIGNED__
	  || 0x7f < c
#else
	  || c < 0
#endif

	)
	ERR_PUSHBACK();
      /* Otherwise, we're appending a character */
      if (buff_end <= buff_p)
	ERR_PUSHBACK();		/* too long! */
      if (c == '\\') {
	result = conn_getc(client_CONN(client), &c);
	if (result != OK)
	  ERR();
	if (c != '\"' && c != '\\')
	  ERR_PUSHBACK();
      }
      *buff_p++ = c;
      break;
    case CLIENT_LITERAL:
      if ('0' <= c && c <= '9') {
	unsigned long   newcount = count * 10 + (c - '0');

	if (newcount < count)
	  ERR_PUSHBACK();	/* overflow */
	/*
	 * XXX This should be fatal if non-synchronizing.
	 */
	count = newcount;
	break;
      }
      if (c == '+') {
	synchronizing = FALSE;
	result = conn_getc(client_CONN(client), &c);
	if (result != OK)
	  ERR();
      }
      if (c != '}')
	ERR_PUSHBACK();
      result = conn_getc(client_CONN(client), &c);
      if (result != OK)
	ERR();
      if (c != '\r')
	ERR_PUSHBACK();
      result = conn_getc(client_CONN(client), &c);
      if (result != OK)
	ERR();
      if (c != '\n')
	ERR_PUSHBACK();
      if (synchronizing) {
	static const char sync_reply[] = "+ \"Ready for data\"\r\n";

	if (client->recovering)
	  return CRLF;
	pthread_mutex_lock(client_OUTPUT_MUTEX(client));
	conn_write(client_CONN(client), sync_reply, sizeof(sync_reply) - 1);
	conn_flush(client_CONN(client));
	pthread_mutex_unlock(client_OUTPUT_MUTEX(client));
      }
      lvalp->str = NULL;
      result = str_alloc(count, NULL, &lvalp->str);
      if (result != OK)
	ERR_PUSHBACK();
      if (client_parser_add_cleanup(client, &free, lvalp->str)) {
	str_free(&lvalp->str);
	ERR_PUSHBACK();
      } {
	char           *it = str_DATA(lvalp->str),
	               *end = it + count;

	while (it < end) {
	  conn_getc(client_CONN(client), it);
	  it++;
	}
      }
      client_SET_STATE(client, CLIENT_NORMAL);
      return STRING;
    case CLIENT_NUMBER:
      if ('0' <= c && c <= '9') {
	unsigned long   newcount = count * 10 + (c - '0');

	if (newcount < count)
	  ERR_PUSHBACK();	/* overflow */
	count = newcount;
      } else {
	lvalp->number = count;
	client_SET_STATE(client, CLIENT_NORMAL);
	conn_PUSHBACK(client_CONN(client), c);
	return NUMBER;
      }
      break;
    case CLIENT_NORMAL:
      if (isalpha((unsigned char) c)) {
	client_SET_STATE(client, CLIENT_ATOM);
	*buff_p++ = tolower(c);
	break;
      }
      switch (c) {
      case '(':
	return '(';
      case ')':
	return ')';
      case ' ':
	return ' ';
      case '\"':
	client_SET_STATE(client, CLIENT_QSTR);
	break;
      case '*':
	return '*';
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
	count = c - '0';
	client_SET_STATE(client, CLIENT_NUMBER);
	break;
      case '{':
	count = 0;
	synchronizing = TRUE;
	client_SET_STATE(client, CLIENT_LITERAL);
	break;
      case '\r':
	client_SET_STATE(client, CLIENT_CR);
	break;
      default:
	ERR_PUSHBACK();
      }
      break;
    case CLIENT_ATOM:
      if (!isalpha((unsigned char) c)) {
	const acap_token_t *token;

	/* We've got the atom. */
	token = acap_token_lookup(buffer, buff_p - buffer);
	if (token) {
	  client_SET_STATE(client, CLIENT_NORMAL);
	  conn_PUSHBACK(client_CONN(client), c);
	  return token->value;
	} else
	  ERR_PUSHBACK();
      }
      if (buff_end <= buff_p)
	ERR_PUSHBACK();		/* atom too long */
      *buff_p++ = tolower(c);
      break;
    }
  }
}
