/* respcode.c--response code implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define RESPCODE_C
#include <config.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "respcode.h"
#include "client.h"
#include "acap_err.h"
#include "../daemon/daemon.h"
#include "../daemon/daemon_err.h"

typedef enum {
  RESP_NOEXIST,
  RESP_WAYTOOMANY,
  RESP_TOOMANY,
  RESP_REFER
}               respcode_tag_t;

struct respcode_s {
  respcode_tag_t  tag;
  union {
    str_t          *str;
    uint32_t        num;
    str_t         **strs;
  }               u;
};

#define respcode_TAG(c) ((c)->tag)

result_t
respcode_write(respcode_t * code,
	       struct client_s * client,
	       result_t *command_result)
{
  result_t        result = OK;

  I(code);
  I(client);
  I(command_result);

  switch (respcode_TAG(code)) {
  case RESP_NOEXIST:
    I(code->u.str);
    result = client_printf(client, " (NoExist %s)",
			   code->u.str);
    break;
  case RESP_WAYTOOMANY:
    result = client_printf(client, " (WayTooMany)");
    break;
  case RESP_TOOMANY:
    result = client_printf(client, " (TooMany %u)",
			   code->u.num);
    if (*command_result == OK)
      *command_result = ACAP_ERR_LIMIT_EXCEEDED;
    break;
  case RESP_REFER:
    {
      str_t **str;
      I(code->u.strs);
      I(code->u.strs[0]);
      result = client_printf(client, " (Refer");
      for (str = code->u.strs;
	   result == OK && *str;
	   ++str)
	result = client_printf(client, " %s", *str);
      if (result == OK)
	result = client_printf(client, ")");
      break;
    }
  default:
    daemon_fatal(DAEMON_ERR_INTERNAL,
		 "Illegal response code type encountered");
  }

  return result;
}

result_t
respcode_free(respcode_t ** code)
{
  I(code);
  I(*code);

  switch (respcode_TAG(*code)) {
  case RESP_NOEXIST:
    I((*code)->u.str);
    str_free(&(*code)->u.str);
    break;
  case RESP_WAYTOOMANY:
  case RESP_TOOMANY:
    break;
  case RESP_REFER:
    I((*code)->u.strs);
    free((*code)->u.strs);
    break;
  default:
    daemon_fatal(DAEMON_ERR_INTERNAL,
		 "Illegal response code type encountered");
  }

  free(*code);
  *code = NULL;

  return OK;
}

static result_t
respcode_alloc(respcode_tag_t tag,
	       respcode_t ** dest)
{
  *dest = malloc(sizeof(respcode_t));
  if (!*dest)
    return errno;

  memset(*dest, 0, sizeof(respcode_t));

  (*dest)->tag = tag;

  return OK;
}

result_t
respcode_noexist_alloc(str_t * path,
		       respcode_t ** dest)
{
  result_t        result;

  result = respcode_alloc(RESP_NOEXIST, dest);
  if (result != OK)
    return result;

  (*dest)->u.str = path;

  return OK;
}

/* This allocates a WAYTOOMANY response code. */
result_t
respcode_waytoomany_alloc(respcode_t ** dest)
{
  return respcode_alloc(RESP_WAYTOOMANY, dest);
}

result_t
respcode_toomany_alloc(unsigned n_entries,
		       respcode_t ** dest)
{
  result_t        result;

  result = respcode_alloc(RESP_TOOMANY, dest);
  if (result != OK)
    return result;

  (*dest)->u.num = n_entries;

  return OK;
}

result_t
respcode_refer_alloc(const strlist_t * strlist,
		     respcode_t ** dest)
{
  result_t result;
  unsigned len;
  strlist_iterator_t it;
  char *data;
  str_t ** strptr;

  I(strlist);
  I(dest);

  result = respcode_alloc(RESP_REFER, dest);
  if (result != OK) return result;

  /* Okay.  How much data will we need? */
  len = sizeof(str_t *);	/* null at end */
  for (strlist_ITER_BEGIN(strlist, &it);
       ! strlist_ITER_ATEND(&it);
       strlist_ITER_INCR(&it))
    len += (sizeof(str_t *)	/* for the pointer */
	    + str_MEM_NEEDED(str_LEN(strlist_ITER_VALUE(&it))));

  /* Allocate that memory... */
  (*dest)->u.strs = malloc(len);
  if (! (*dest)->u.strs) {
    free(*dest);
    *dest = NULL;
    return errno;
  }

  /* And fill it in. */
  strptr = (*dest)->u.strs;
  data = (char *)(strptr + 1 + strlist_N_VALUES(strlist));

  *strptr++ = (str_t *) data;

  for (strlist_ITER_BEGIN(strlist, &it);
       ! strlist_ITER_ATEND(&it);
       strlist_ITER_INCR(&it)) {
    *strptr++ = (str_t *) data;
    memcpy(data,
	   strlist_ITER_VALUE(&it),
	   str_MEM_NEEDED(str_LEN(strlist_ITER_VALUE(&it))));
    data += str_MEM_NEEDED(str_LEN(strlist_ITER_VALUE(&it)));
  }

  *strptr = NULL;		/* termination */
  return OK;
}
