/* store_ast.h--store AST definitions
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef STORE_AST_H
#define STORE_AST_H

#include "../util/sllist.h"
#include "../util/str.h"
#include "attr.h"
#include "stamp.h"

struct client_s;

typedef attr_state_t store_metadatum_state_t;

typedef enum {
  METADATUM_ACL,
  METADATUM_VALUE,
  METADATUM_CONTENT_TYPE,
  METADATUM_UNNAMED,
  METADATUM_MAX_METADATUM = METADATUM_UNNAMED
} store_metadatum_name_t;

typedef struct store_metadatum_s store_metadatum_t;

typedef struct store_attribute_s store_attribute_t;

typedef struct store_params_s store_params_t;

typedef struct store_entry_s store_entry_t;

typedef struct store_s store_t;

struct store_metadatum_s {
  store_metadatum_name_t name;
  store_metadatum_state_t state;
  sllist_t       *data;
};

typedef sllist_iterator_t store_metadatum_iterator_t;

#define store_metadatum_NAME(m) ((m)->name)
#define store_metadatum_STATE(m) ((m)->state)
#define store_metadatum_VALUE(m) ((str_t *)(sllist_DATA((m)->data)))

#define store_metadatum_ITER_BEGIN(m, i) sllist_ITER_BEGIN((m)->data, i)
#define store_metadatum_ITER_ATEND(i) sllist_ITER_ATEND(i)
#define store_metadatum_ITER_INCR(i) sllist_ITER_INCR(i)
#define store_metadatum_ITER_VALUE(i) \
	((str_t *)sllist_DATA(sllist_ITER_VALUE(i)))

struct store_attribute_s {
  str_t          *name;
  store_metadatum_t *values[METADATUM_MAX_METADATUM];
};

#define store_attribute_NAME(a) ((a)->name)
#define store_attribute_METADATUM(a, i) ((a)->values[i])

struct store_params_s {
  bool_t          nocreate;
  str_t          *unchangedsince;
};

struct store_entry_s {
  store_entry_t  *next;
  str_t          *path;
  str_t          *name;
  store_params_t *params;
  store_attribute_t *entry_attribute;
  store_metadatum_t *inherit_value;
  sllist_t       *attributes;
};

typedef sllist_iterator_t store_entry_iterator_t;

/* path is the entire path to the entry;
 * name is the specific entry name. */
#define store_entry_PATH(e) ((e)->path)
#define store_entry_NAME(e) ((e)->name)
#define store_entry_UNCHANGEDSINCE(e) ((e)->params->unchangedsince)
#define store_entry_CREATE(e) (! (e)->params->nocreate)
#define store_entry_ENTRY_ATTR(e) ((e)->entry_attribute)
#define store_entry_INHERIT_VALUE(e) ((e)->inherit_value)

#define store_entry_ITER_BEGIN(e, i) sllist_ITER_BEGIN((e)->attributes, i)
#define store_entry_ITER_ATEND(i) sllist_ITER_ATEND(i)
#define store_entry_ITER_INCR(i) sllist_ITER_INCR(i)
#define store_entry_ITER_VALUE(i) ((store_attribute_t *)		\
				   sllist_DATA(sllist_ITER_VALUE(i)))

struct store_s {
  store_entry_t  *entries;
  str_t          *tag;
  struct client_s *client;
};

typedef struct {
  store_entry_t  *ent;
}               store_iterator_t;

#define store_ITER_BEGIN(s, i) ((i)->ent = (s)->entries)
#define store_ITER_ATEND(i) ((i)->ent == NULL)
#define store_ITER_INCR(i) ((i)->ent = (i)->ent ? (i)->ent->next : NULL)
#define store_ITER_VALUE(i) ((i)->ent)

#define store_TAG(s) ((s)->tag)
#define store_CLIENT(s) ((s)->client)

extern          result_t
store_alloc(struct client_s * client,
	    store_entry_t * first_entry,
	    store_t ** dest);

extern result_t store_free(store_t ** store);

extern result_t
store_append(store_t * store,
	     struct client_s * client,
	     store_entry_t * entry);

extern result_t
store_params_alloc(struct client_s * client,
		   store_params_t ** dest);

extern result_t store_params_set_nocreate(store_params_t * params);

extern result_t
store_params_set_unchangedsince(store_params_t * params,
				str_t * unchangedsince);

extern result_t
store_entry_alloc(struct client_s * client,
		  store_attribute_t * first_attribute,
		  store_entry_t ** dest);

extern result_t
store_entry_set_path(store_entry_t * entry,
		     struct client_s * client,
		     str_t * path);

/* Note: this makes a copy of the name;
 * caller is responsible for freeing the original
 * (a little wacky, because we're not doing this
 * while parsing). */
extern result_t
store_entry_set_name(store_entry_t * entry,
		     str_t * name);

extern result_t
store_entry_set_params(store_entry_t * entry,
		       store_params_t * params);

extern result_t
store_entry_set_unchangedsince(store_entry_t * entry,
			       struct client_s * client,
			       str_t * unchangedsince);

extern result_t
                store_entry_set_nocreate(store_entry_t * entry);

extern result_t
store_entry_append(store_entry_t * entry,
		   struct client_s * client,
		   store_attribute_t * attribute);

/* This one gets called by the store code *after* the transaction
 * has begun; we need to do it then, because we don't know the modtime
 * until we're actually processing the request.
 */
extern result_t
store_entry_complete(store_entry_t * entry,
		     const str_t * modtime);

extern result_t
store_attribute_alloc(struct client_s * client,
		      store_metadatum_t * first_value,
		      store_attribute_t ** dest);

extern result_t
store_attribute_value_alloc(struct client_s * client,
			    str_t * value,
			    store_attribute_t ** dest);

extern result_t
store_attribute_set_name(store_attribute_t * attr,
			 struct client_s * client,
			 str_t * name);

extern result_t
store_attribute_append(store_attribute_t * attr,
		       struct client_s * client,
		       store_metadatum_t * value);

/* A metadatum's "name" here is its lexical token; see parse.acap.h. */
extern result_t
store_metadatum_set_name(store_metadatum_t * meta,
			 struct client_s * client,
			 int name);

extern result_t
store_metadatum_str_alloc(struct client_s * client,
			  str_t * data,
			  store_metadatum_t ** dest);

/* This one isn't actually used by the parser;
 * it's used to rewrite metadata which've been defaulted,
 * filling in their correct values. */
extern result_t
store_metadatum_set_str(store_metadatum_t * metadatum,
			str_t * data);

extern result_t
store_metadatum_nil_alloc(struct client_s * client,
			  store_metadatum_t ** dest);

extern result_t
store_metadatum_default_alloc(struct client_s * client,
			      store_metadatum_t ** dest);

extern result_t
store_metadatum_multi_alloc(struct client_s * client,
			    str_t * first_value,
			    store_metadatum_t ** dest);

extern result_t
store_metadatum_multi_append(store_metadatum_t * meta,
			     struct client_s * client,
			     str_t * value);

#endif				/* STORE_AST_H */
