/* trx.c--transaction implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

/* Nothing magical, here--DB does all the dirty work. */

#define TRX_C
#include <config.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>
#include <db.h>
#include "trx.h"
#include "acap_err.h"
#include "entry_ref.h"
#include "entry_ref_history.h"
#include "../util/util_err.h"

struct trx_cleanup_s {
  trx_cleanup_t  *next;
  trx_cleanup_proc_t *proc;
  void           *rock;
};

/* Yeah, static globals are evil, and this really ought
 * to be coded up as a singleton.  Deal. */
static pthread_mutex_t trx_mutex = PTHREAD_MUTEX_INITIALIZER;

static stamp_t  trx_stamp = STAMP_ZERO_INITIALIZER;

static trx_t   *trx_oldest = NULL,
               *trx_newest = NULL;

result_t
trx_begin(trx_t ** dest,
	  trx_type_t trx_type)
{
  result_t        result;

  time_t          now;

  I(dest);
  I(acapd_db_env->tx_info);

  *dest = (trx_t *) malloc(sizeof(trx_t));
  if (!*dest)
    return errno;

  result = pthread_mutex_init(trx_MUTEX(*dest), NULL);
  if (result != OK)
    goto cleanup_dest;

  result = hash_init(&hash_ptr_procs, trx_ENTRY_REF_LOCKS(*dest));
  if (result != OK)
    goto cleanup_mutex;

  (*dest)->next = NULL;
  (*dest)->in_use = TRUE;
  (*dest)->immediate_cleanup = NULL;
  (*dest)->final_cleanup = NULL;

  (*dest)->type = trx_type;

  pthread_mutex_lock(&trx_mutex);

  result = txn_begin(acapd_db_env->tx_info,
		     NULL,
		     &(*dest)->txn);

  if (result < 0)
    result = ACAP_ERR_DB;

  if (result != OK)
    goto cleanup_hash;

  if (trx_newest) {
    I(!trx_newest->next);
    trx_newest->next = *dest;
  }
  trx_newest = *dest;
  if (!trx_oldest)
    trx_oldest = *dest;

  now = time(NULL);
  if (stamp_SEC(&trx_stamp) < now)
    stamp_SET(&trx_stamp, now, 0);
  else
    stamp_INCR_FRAC(&trx_stamp);

  stamp_COPY(&(*dest)->stamp, &trx_stamp);

  if (trx_type == TRX_RO)
    pthread_mutex_unlock(&trx_mutex);

  return result;

cleanup_hash:
  hash_destroy(trx_ENTRY_REF_LOCKS(*dest));

cleanup_mutex:
  pthread_mutex_destroy(trx_MUTEX(*dest));

cleanup_dest:
  free(*dest);
  *dest = NULL;

  return result;
}

result_t
trx_drop_lock(trx_t * trx)
{
  I(trx);
  I(trx->type == TRX_RW);

  pthread_mutex_unlock(&trx_mutex);
  return OK;
}

static void
trx_run_cleanups(trx_t * trx,
		 trx_cleanup_t * cleanup)
{
  while (cleanup) {
    trx_cleanup_t  *next = cleanup->next;

    cleanup->proc(trx->commit_state, cleanup->rock);
    free(cleanup);
    cleanup = next;
  }
}

static result_t
trx_cleanup_locks(trx_commit_t commit_state,
		  trx_t * trx)
{
  result_t        result;

  hash_iter_t     it;

  hash_entry_t   *ent;

  entry_ref_t    *entry_ref;

  I(trx);

  VL(("Cleaning up locks of entries changed by trx %p\n", trx));

  result = hash_iter_begin(trx_ENTRY_REF_LOCKS(trx), &it);
  if (result != OK)
    return result;

  while ((result = hash_iter_next(&it, &ent)) == OK) {
    entry_ref = *(entry_ref_t **) hash_GET_KEY(ent);

    VL(("Cleaning up lock on entry_ref %p\n", entry_ref));

    pthread_mutex_lock(entry_ref_MUTEX(entry_ref));
    I(entry_ref_WRITER(entry_ref));
    if (commit_state == TRX_COMMIT) {
      entry_ref_history_SET_NEXT(entry_ref_WRITER(entry_ref),
				 entry_ref_HIST(entry_ref));
      entry_ref_HIST(entry_ref) = entry_ref_WRITER(entry_ref);
      entry_ref_WRITER(entry_ref) = NULL;
    } else {
      entry_ref_history_free(&entry_ref_WRITER(entry_ref));
    }
    pthread_cond_broadcast(entry_ref_COND(entry_ref));
    pthread_mutex_unlock(entry_ref_MUTEX(entry_ref));
  }

  if (result == HASH_ITER_END)
    result = OK;

  return result;
}

static result_t
trx_cleanup_entries(trx_t * trx)
{
  /* No data older than this trx needs to be kept around now. */
  result_t        result;

  hash_iter_t     it;

  hash_entry_t   *ent;

  entry_ref_t    *entry_ref;

  entry_ref_history_t *history,
                 *next;

  I(trx);

  VL(("Cleaning up histories of entries changed by trx %p\n", trx));

  result = hash_iter_begin(trx_ENTRY_REF_LOCKS(trx), &it);
  if (result != OK)
    return result;

  while ((result = hash_iter_next(&it, &ent)) == OK) {
    entry_ref = *(entry_ref_t **) hash_GET_KEY(ent);

    VL(("Cleaning up history of entry_ref %p\n", entry_ref));

    pthread_mutex_lock(entry_ref_MUTEX(entry_ref));

    I(entry_ref_HIST(entry_ref));

    /*
     * Note that this won't catch a history entry that's younger than the
     * current trx if that history entry is the first entry; this is
     * intentional, because the first item in the history is always valid.
     */
    for (history = entry_ref_HIST(entry_ref);
	 (entry_ref_history_NEXT(history)
	  && stamp_LT(trx_STAMP(trx),
		 entry_ref_history_STAMP(entry_ref_history_NEXT(history))));
	 history = entry_ref_history_NEXT(history)) {
    }

    /* We want to delete next, and everything following it... */
    next = entry_ref_history_NEXT(history);
    entry_ref_history_NEXT(history) = NULL;
    history = next;
    while (history) {
      next = entry_ref_history_NEXT(history);
      entry_ref_history_free(&history);
      history = next;
    }

    pthread_mutex_unlock(entry_ref_MUTEX(entry_ref));
    entry_ref_unref(&entry_ref);
  }
  if (result == HASH_ITER_END)
    result = OK;

  return result;
}

static result_t
trx_free(trx_t ** trx)
{
  I(trx);
  I(*trx);
  I((*trx)->in_use == TRX_COMPLETE);

  trx_cleanup_entries(*trx);

  pthread_mutex_destroy(trx_MUTEX(*trx));

  hash_destroy(trx_ENTRY_REF_LOCKS(*trx));

  trx_run_cleanups(*trx, (*trx)->final_cleanup);

  free(*trx);
  *trx = NULL;

  return OK;
}

result_t
trx_done(trx_t ** trx,
	 trx_commit_t commit)
{
  result_t        result = OK;

  I(trx);
  I(*trx);
  I((*trx)->in_use == TRX_ACTIVE);
  I(trx_oldest);
  I(trx_newest);

  (*trx)->commit_state = commit;

  trx_run_cleanups(*trx, (*trx)->immediate_cleanup);

  if (trx_DB_TXN(*trx)) {
    if (commit == TRX_COMMIT)
      result = txn_commit(trx_DB_TXN(*trx));
    else
      result = txn_commit(trx_DB_TXN(*trx));	/* TODO: abort */
    if (result < 0)
      result = ACAP_ERR_DB;
  }
  trx_cleanup_locks(commit, *trx);

  pthread_mutex_lock(&trx_mutex);

  (*trx)->in_use = TRX_COMPLETE;
  while (trx_oldest && trx_oldest->in_use == TRX_COMPLETE) {
    trx_t          *next = trx_oldest->next;

    I(!next || stamp_LT(trx_STAMP(trx_oldest),
			trx_STAMP(next)));
    trx_free(&trx_oldest);
    trx_oldest = next;
    if (!trx_oldest)
      trx_newest = NULL;
  }

  pthread_mutex_unlock(&trx_mutex);

  *trx = NULL;

  return result;
}

result_t
trx_schedule_cleanup(trx_t * trx,
		     trx_cleanup_time_t cleanup_time,
		     trx_cleanup_proc_t * proc,
		     void *rock)
{
  trx_cleanup_t  *cleanup;

  I(trx);
  I(proc);
  I(rock);
  I(trx_DB_TXN(trx));

  cleanup = (trx_cleanup_t *) malloc(sizeof(trx_cleanup_t));
  if (!cleanup)
    return errno;

  cleanup->proc = proc;
  cleanup->rock = rock;

  pthread_mutex_lock(&trx->mutex);

  if (cleanup_time == TRX_CLEANUP_IMMEDIATE) {
    cleanup->next = trx->immediate_cleanup;
    trx->immediate_cleanup = cleanup;
  } else {
    cleanup->next = trx->final_cleanup;
    trx->final_cleanup = cleanup;
  }

  pthread_mutex_unlock(&trx->mutex);

  return OK;
}

result_t
trx_str(trx_t * trx,
	str_t ** str)
{
  char            buf[200];	/* this had BETTER be longer than the longest
				 * timestamp... if not, I'm officially
				 * scared. */

  result_t        result = OK;

  struct tm       gmt;

  time_t          sec;

  I(trx);
  I(str);

  sec = stamp_SEC(trx_STAMP(trx));
  gmtime_r(&sec, &gmt);

  sprintf(buf, "%04d%02d%02d%02d%02d%02d%010u",
	  gmt.tm_year + 1900,
	  gmt.tm_mon + 1,
	  gmt.tm_mday,
	  gmt.tm_hour,
	  gmt.tm_min,
	  gmt.tm_sec,
	  stamp_FRAC(trx_STAMP(trx)));

  result = str_copy_cstr(buf, str);

  return result;
}

result_t
trx_lock_entry(trx_t * trx,
	       entry_ref_t * entry_ref)
{
  result_t        result = OK;

  entry_ref_history_t *history;

  I(trx);
  I(trx_IS_RW(trx));
  I(entry_ref);

  pthread_mutex_lock(entry_ref_MUTEX(entry_ref));

  if (!entry_ref_WRITER(entry_ref)) {
    hash_entry_t   *hashent;

    hash_did_insert_t inserted;

    VL(("Adding changed entry_ref %p to trx %p...\n",
	entry_ref, trx));

    pthread_mutex_lock(trx_MUTEX(trx));
    result = hash_find(trx_ENTRY_REF_LOCKS(trx),
		       entry_ref,
		       HASH_INSERT,
		       &inserted,
		       &hashent);
    I(inserted == HASH_DID_INSERT);
    pthread_mutex_unlock(trx_MUTEX(trx));
    if (result == OK) {
      history = entry_ref_HIST(entry_ref);
      I(history);

      result = entry_ref_history_alloc(*trx_STAMP(trx),
				       entry_ref_history_DBPATH(history),
				       entry_ref_history_DBKEY(history),
				       &entry_ref_WRITER(entry_ref));
      if (result == OK)
	entry_ref_ref(entry_ref);	/* trx now has a ref on it */
      else
	hash_delete_entry(hashent);
    }
  }
  pthread_mutex_unlock(entry_ref_MUTEX(entry_ref));

  return result;
}
