/* trx.h--transaction interface
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef TRX_H
#define TRX_H

/* Things are used to keep track of things which shouldn't go away;
 * for instance, even when an entry's deleted, a search currently in
 * progress might need that entry (it's darn near impossible to tell,
 * without keeping lots of data resident in memory and throwing a
 * global lock around it all).  So, commands which can't immediately
 * tell what they'll be touching allocate things, and commands which
 * modify things allocate things, and cleanup actions are hung on
 * modification things, and when a thing's the oldest thing on the
 * system and its command has completed, it can tell everything it was
 * keeping alive to actually perform cleanups.  Which is why this
 * doesn't quite follow the alloc()/free() convention--things are
 * freed out of band. */

#include <pthread.h>
#include "stamp.h"
#include "../util/str.h"
#include "../util/hash.h"

struct entry_ref_s;

typedef struct trx_cleanup_s trx_cleanup_t;

typedef enum {
  TRX_RO,
  TRX_RW
}               trx_type_t;

typedef enum {
  TRX_ABORT,
  TRX_COMMIT
}               trx_commit_t;

typedef struct trx_s {
  struct trx_s   *next;
  pthread_mutex_t mutex;
  stamp_t         stamp;
  enum {
    TRX_COMPLETE,
    TRX_ACTIVE
  }               in_use;
  trx_type_t      type;
  trx_cleanup_t  *immediate_cleanup,
                 *final_cleanup;
  trx_commit_t    commit_state;
  struct __db_txn *txn;
  hash_table_t    entry_ref_locks;
}               trx_t;

#define trx_DB_TXN(t) ((t) ? (DB_TXN *)(t)->txn : NULL)
#define trx_STAMP(t) (&(t)->stamp)
#define trx_IS_RW(t) ((t)->type == TRX_RW)
#define trx_MUTEX(t) (&(t)->mutex)
#define trx_ENTRY_REF_LOCKS(t) (&(t)->entry_ref_locks)

/* Instead of the standard alloc/free interface, I'm using a slightly
 * different style for this; you create a transaction with
 * trx_begin(), and complete it with either trx_commit() or
 * trx_rollback(). */

/* Note that every trx_begin() must be completed by a trx_done(),
 * and that a global lock will be held by RW transactions until
 * trx_drop_lock() is called; trx_drop_lock() *must* be called
 * before trx_done() for RW transactions. */

extern          result_t
trx_begin(trx_t ** dest,
	  trx_type_t trx_type);

/* trx_drop_lock should be called by writers which have obtained their
 * locks; this allows readers to run. */
extern result_t
                trx_drop_lock(trx_t * trx);

extern result_t
trx_done(trx_t ** trx,
	 trx_commit_t commit);

typedef void
trx_cleanup_proc_t(trx_commit_t committed,
		   void *rock);

/* The cleanup proc here will be called with the specified rock
 * when the thing is the oldest thing on the system, and when the
 * thing's associated command has indicated that it's done with
 * the thing. */

typedef enum {
  TRX_CLEANUP_IMMEDIATE,
  TRX_CLEANUP_FINAL
}               trx_cleanup_time_t;

extern result_t
trx_schedule_cleanup(trx_t * thing,
		     trx_cleanup_time_t cleanup_time,
		     trx_cleanup_proc_t * proc,
		     void *rock);

/* This can be used to get the ACAP modtime corresponding to this
 * thing's timestamp; this is quite useful when logging, generating
 * modtimes for entries, etc... */
extern result_t
trx_str(trx_t * trx,
	str_t ** str);

/* This adds the entry to the trx's pool of locked entries, creating a
 * writer history field for it in the process (initializing it to be a
 * duplicate of the existing history field).  This assumes that the
 * entry_ref is unlocked. */
extern result_t
trx_lock_entry(trx_t * trx,
	       struct entry_ref_s * entry_ref);

#endif				/* TRX_H */
