/* conn.c - connection implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define CONN_C
#include <config.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/time.h>

#ifdef HAVE_SYS_FILIO
#include <sys/filio.h>
#endif

#include <syslog.h>
#include <errno.h>
#include <com_err.h>
#include <fcntl.h>
#include "conn.h"
#include "daemon_err.h"
#include "io.h"
#include "dispatch.h"
#include "daemon.h"

#undef L_DEFAULT_GUARD
#define L_DEFAULT_GUARD (debug_daemon_io)

static unsigned conn_handler_concurrency = CONN_HANDLER_DEFAULT_CONCURRENCY;

static dispatch_queue_t *conn_handler_queue = NULL;

static pthread_once_t conn_handler_queue_once = PTHREAD_ONCE_INIT;

static void
conn_handler_queue_init(void)
{
  result_t result;

  result = dispatch_queue_alloc(&conn_handler_queue,
				conn_handler_concurrency);

  if (result)
    daemon_fatal(result, "Initializing connection dispatch queue");
}

result_t
conn_set_concurrency(unsigned concurrency)
{
  conn_handler_concurrency = concurrency;
  return OK;
}

/* Notes:
 * When IO comes in, it'll go to conn_data_handler.
 * If there's no thread running, conn_data_handler will queue the connection
 * for processing via conn_data_loop.
 * If there *is* a thread running, conn_data_handler won't start one. :-)
 * The conn_data_loop attempts to read a command.
 * Afterwards, if there's no data available, it'll punt;
 * if there *is* some, it'll try to read a command...
 * The lexer might read a partial command, and want more.
 * It'll call conn_moredata.
 * conn_moredata will simply do a blocking read,
 * because the parser keeps its state on the stack; it can't just be restarted.
 * Simple, eh?
 */

static          bool_t
conn_is_fd_readable(conn_t * conn)
{
  /*
   * We want to try to parse iff a read won't block (if a read *will* block,
   * then as soon as it won't, we'll get a SIGPOLL and another handler
   * thread).  How do you test that?  A polling select() seems to be the
   * trick...
   */

  struct timeval  timeout;

  fd_set          fds;

  FD_ZERO(&fds);
  FD_SET(conn->fd, &fds);
  timeout.tv_sec = 0;
  timeout.tv_usec = 0;
  if (!select(conn->fd + 1, &fds, NULL, NULL, &timeout))
    return FALSE;		/* It's not readable. */
  else
    return TRUE;		/* It *is* readable */
}

static void
conn_data_loop(conn_t * conn)
{
  /* Entry point for threads processing connections. */
  bool_t          done = FALSE;

  VL(("Starting connection processing thread for conn %p\n", conn));

  conn->input.buffer_position
    = conn->input.buffer_end
    = conn->input.buffer
    = NULL;

  while (!done) {
    bool_t          connection_closed;

    VL(("Running parser function for conn %p\n", conn));

    (*conn->parse_func) (conn->parse_rock, &connection_closed);

    if (connection_closed) {	/* we're done */
      VL(("Connection has been closed; relinquishing processor thread\n"));
      return;
    }
    /* Find out if there's any more data... */

    /*
     * Note that we *must* hold this mutex between the time when we check for
     * more data and the time when we set conn->active to FALSE; this means
     * that no matter when data comes in, either we will see it (and keep
     * processing), or we won't, and we'll set conn->active to FALSE before
     * anything else can check whether or not we'll be dealing with it.
     */
    pthread_mutex_lock(&conn->mutex);

    if (conn->input.buffer_position
	< conn->input.buffer_end) {	/* check the buffer */
      done = FALSE;		/* we've already read some data; we're set */
    } else {
      /*
       * Otherwise, we have to do more work.  Bummer.
       */
      done = !conn_is_fd_readable(conn);
    }

    if (done)
      conn->active = FALSE;

    pthread_mutex_unlock(&conn->mutex);
  }				/* while(!done) */

  /*
   * Well... we don't think there's any more data. Let this thread go do
   * something useful.
   */
  conn_flush(conn);

  /* A little cleanup... */
  if (conn->input.buffer)
    free(conn->input.buffer);

  conn->input.buffer_position
    = conn->input.buffer_end
    = conn->input.buffer
    = NULL;

  VL(("Relinquishing thread for conn %p\n", conn));
}

static void
                conn_data_handler(int fd __attribute__((unused)),
				                  conn_t * conn) {
  /*
   * This is called when a SIGPOLL has occurred and the connection is
   * readable.  If there's no thread processing this connection, we need to
   * queue it to be processed.
   */
  I(conn);

  VL(("Data pending for conn %p\n", conn));

  pthread_mutex_lock(&conn->mutex);

  if (!conn->active) {
    /* It needs a thread to process the data */
    result_t        result;

    pthread_once(&conn_handler_queue_once,
		 &conn_handler_queue_init);

    VL(("Queueing thread request for conn %p\n", conn));

    result = dispatch_enqueue(conn_handler_queue,
			      (dispatch_task_t *) & conn_data_loop,
			      conn);
    if (result) {
      /*
       * Log the failure, somehow - there's really very little we can do to
       * recover from this, unfortunately.
       */
      VL(("Fatal error: Failed to enqueue a connection handler"));
      abort();
    } else {
      conn->active = TRUE;
    }
  } else {
    /*
     * There's already an active processing thread - so we don't have
     * anything to do.
     */
  }
  pthread_mutex_unlock(&conn->mutex);
}

result_t
conn_alloc(conn_t ** conn,
	   struct sockaddr_in * peer,
	   int fd,
	   conn_parser_t * parse_func,
	   void *parse_rock)
{
  result_t        result;

  I(conn);
  I(!*conn);
  I(0 <= fd);

  *conn = (conn_t *) malloc(sizeof(conn_t));
  if (!*conn)
    return errno;

  result = pthread_mutex_init(&(*conn)->mutex, NULL);
  if (result)
    goto cleanup_conn;

  (*conn)->active = FALSE;
  (*conn)->shutdown = FALSE;
  (*conn)->fd = fd;
  (*conn)->peer = *peer;
  (*conn)->parse_func = parse_func;
  (*conn)->parse_rock = parse_rock;
  (*conn)->input.buffer_position
    = (*conn)->input.buffer_end
    = (*conn)->input.buffer
    = (*conn)->output.buffer_position
    = (*conn)->output.buffer_end
    = (*conn)->output.buffer
    = NULL;
  (*conn)->sasl = NULL;
  (*conn)->mechname = NULL;

  result = io_add_handler(fd, (io_handler_t *) & conn_data_handler, *conn);
  if (result)
    goto cleanup_mutex;

  VL(("Allocated conn %p\n", *conn));

  /* There might *already* be data pending on the socket... */
  if (conn_is_fd_readable(*conn))
    conn_data_handler(fd, *conn);

  return OK;

cleanup_mutex:
  pthread_mutex_destroy(&(*conn)->mutex);

cleanup_conn:
  free(*conn);
  *conn = NULL;

  return result;
}

result_t
conn_set_sasl(conn_t * conn,
	      sasl_conn_t * sasl)
{
  result_t        result;

  char           *output_buffer;

  size_t          output_len = 0;

  sasl_ssf_t      ssf;

  I(conn);
  I(sasl);
  I(!conn->sasl);

  {
    const char     *username,
                   *username_end;

    uint32_t        addr = conn->peer.sin_addr.s_addr;

    if (sasl_getprop(sasl, SASL_USERNAME, (void **) &username) != SASL_OK)
      username = "(unknown)";

    for (username_end = username;
	 *username_end;
	 username_end++)
      if (iscntrl((unsigned char) *username_end))
	break;

    syslog(LOG_INFO, "login: [%d.%d.%d.%d] %*s %s",
	   (addr >> 24) & 0xFF,
	   (addr >> 16) & 0xFF,
	   (addr >> 8) & 0xFF,
	   (addr) & 0xFF,
	   username_end - username,
	   username,
	   str_DATA(conn->mechname));
  }

  sasl_getprop(sasl, SASL_SSF, (void **) &ssf);
  VL(("Setting SASL security layer; ssf is %d\n", ssf));

  if (ssf) {
    /*
     * (If the SSF is 0, there's no reason to run everything through the SASL
     * layer; we can just send the data directly.)
     */

    /* Flush the output side... */
    result = conn_flush(conn);
    if (result)
      return result;

    /*
     * And then decode all pending input (which is presumably encoded)
     */
    if (conn->input.buffer_position != conn->input.buffer_end) {
      I(conn->input.buffer);

      if (sasl_decode(sasl,
		      conn->input.buffer_position,
		      conn->input.buffer_end - conn->input.buffer_position,
		      &output_buffer, &output_len) != SASL_OK) {
	conn->shutdown = TRUE;	/* !give!up */
	return DAEMON_ERR_SASL;
      }
      I(conn->input.buffer_end - conn->input.buffer_position
	== (ssize_t) output_len);
      free(conn->input.buffer);
      conn->input.buffer = output_buffer;
      conn->input.buffer_position = output_buffer;
      conn->input.buffer_end = output_buffer + output_len;
    }
    /* ... and set the conn. */
    conn->sasl = sasl;
  }
  return OK;
}

result_t
_conn_moredata(conn_t * conn, char *c)
{
  /* Gets more data from the connection. */
  result_t        result = OK;

  ssize_t         n_read;

  I(conn);
  I(c);
  I(conn->input.buffer_position == conn->input.buffer_end);

  VL(("Reading data from conn %p\n", conn));

  if (conn->shutdown) {
    /*
     * This connection's toast; we'll never read anything. Time to punt back
     * to the top...
     */
    VL(("Conn %p is shut down - returning EOF\n", conn));
    return DAEMON_ERR_EOF;
  }
  if (conn->input.buffer) {
    free(conn->input.buffer);
    conn->input.buffer_position
      = conn->input.buffer_end
      = conn->input.buffer
      = NULL;
  }
  conn->input.buffer = malloc(CONN_BUFFER_SIZE);
  if (!conn->input.buffer)
    return errno;
  /*
   * This'll block for IO... unfortunate, but necessary, since bison's
   * keeping everything on the stack.
   */

try_to_read_more:
  VL(("Reading %d octets from fd %d\n", CONN_BUFFER_SIZE, conn->fd));

  n_read = read(conn->fd, conn->input.buffer, CONN_BUFFER_SIZE);

  VL(("Read %d octets from fd %d\n", n_read, conn->fd));

  if (n_read < 0) {
    result = errno;
    if (result == OK) {
      result = DAEMON_ERR_EOF;
      conn->shutdown = TRUE;
    }
    return result;
  }
  if (n_read == 0) {
    conn->shutdown = TRUE;	/* connection is Toast */
    return DAEMON_ERR_EOF;
  }
  if (conn->sasl) {
    /* We need to do a SASL decode */
    char           *output_buffer;

    unsigned        output_len;

    if (sasl_decode(conn->sasl, conn->input.buffer, n_read,
		    &output_buffer, &output_len) != SASL_OK) {
      VL(("SASL decode failed\n"));
      conn->shutdown = TRUE;	/* !give!up */
      return DAEMON_ERR_SASL;
    }
    if (output_len == 0) {
      I(!output_buffer);
      goto try_to_read_more;
    }
    I(output_buffer);

    free(conn->input.buffer);
    conn->input.buffer = output_buffer;
    n_read = output_len;
  }
  conn->input.buffer_position = conn->input.buffer;
  conn->input.buffer_end = conn->input.buffer + n_read;

  *c = *conn->input.buffer_position++;

  VL(("After _moredata: input.buffer_position = %p, buffer_end = %p\n", conn->input.buffer_position, conn->input.buffer_end));

  return result;
}

static result_t
_conn_alloc_output_buffer(conn_t * conn)
{
  I(conn);
  I(!conn->output.buffer);

  conn->output.buffer = malloc(CONN_BUFFER_SIZE);
  if (!conn->output.buffer)
    return errno;

  conn->output.buffer_position = conn->output.buffer;
  conn->output.buffer_end = conn->output.buffer + CONN_BUFFER_SIZE;

  return OK;
}

static result_t
_conn_flush(conn_t * conn, bool_t keep_conn_buffer)
{
  result_t        result = OK;

  char           *buffer = NULL;

  size_t          len;

  bool_t          free_buffer;

  I(conn);
  I(conn->output.buffer <= conn->output.buffer_position);

  buffer = conn->output.buffer;
  len = conn->output.buffer_position - buffer;

  if (len && conn->sasl) {
    /* Do a SASL encode */
    if (sasl_encode(conn->sasl, buffer, len,
		    &buffer, &len))
      errno = DAEMON_ERR_SASL;
    free_buffer = TRUE;
  } else
    free_buffer = FALSE;

  if (len && write(conn_FD(conn), buffer, len) != (ssize_t) len) {
    result = errno;
  }
  if (keep_conn_buffer) {
    /* We want to have a buffer coming out of here. */
    if (conn->output.buffer)	/* we've already got one... */
      conn->output.buffer_position = conn->output.buffer;
    else if (result == OK)	/* we want to create one. */
      result = _conn_alloc_output_buffer(conn);
  } else {
    free(conn->output.buffer);
    conn->output.buffer_position
      = conn->output.buffer
      = conn->output.buffer_end
      = NULL;
  }

  if (free_buffer && buffer)
    free(buffer);

  return result;
}

result_t
conn_flush(conn_t * conn)
{
  I(conn);
  return _conn_flush(conn, FALSE);
}

result_t
_conn_flushc(conn_t * conn, char c)
{
  result_t        result;

  I(conn);

  result = _conn_flush(conn, TRUE);
  if (result)
    return result;

  *(conn->output.buffer_position++) = c;
  return OK;
}

result_t
conn_write(conn_t * conn,
	   const char *data,
	   size_t len)
{
  result_t        result;

  I(conn);
  I(data);

  /* Okay, there're two cases: we can buffer it, and we can't... */
  if (!conn->output.buffer && len <= CONN_BUFFER_SIZE) {
    /* It's worth creating a buffer for this. */
    result = _conn_alloc_output_buffer(conn);
    if (result)
      return result;
  }
  if (0 <= (ssize_t) len
      && (ssize_t) len
      <= conn->output.buffer_end - conn->output.buffer_position) {
    /* Yea, we win. */
    memcpy(conn->output.buffer_position, data, len);
    conn->output.buffer_position += len;
  } else {
    /*
     * Otherwise, we can't buffer it.  So let's transfer what we can, and
     * send the rest in one Big write.
     */
    I(len);

    result = _conn_flush(conn, TRUE);
    if (result)
      return result;
    if (conn->sasl) {
      /* Do a SASL encode */
      char           *buffer;

      if (sasl_encode(conn->sasl, data, len,
		      &buffer, &len))
	return DAEMON_ERR_SASL;
      else if (write(conn_FD(conn), buffer, len) != (ssize_t) len) {
	free(buffer);
	return errno;
      }
      free(buffer);
    } else
      /* Do a raw write */
    if (write(conn_FD(conn), data, len) != (ssize_t) len)
      return errno;
  }
  return OK;
}

result_t
conn_free(conn_t ** conn)
{
  I(conn);
  I(*conn);

  VL(("Freeing conn %p\n", *conn));

  if ((*conn)->input.buffer)
    free((*conn)->input.buffer);

  if ((*conn)->output.buffer)
    free((*conn)->output.buffer);

  if ((*conn)->mechname)
    str_free(&(*conn)->mechname);

  io_remove_handler((*conn)->fd);
  close((*conn)->fd);
  free(*conn);
  *conn = NULL;

  return OK;
}
