/* hash.c - hash table implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define HASH_C
#include <config.h>
#include <stdlib.h>
#include <string.h>
#include "hash.h"
#include "str.h"
#include "errno.h"
#include "util_err.h"

/* TODO: HASH_PTR_SHIFT should be based on the alignment of the pointers. */
#define HASH_PTR_SHIFT (3)
#define HASH_BUCKET_MULTIPLIER (3)
#define HASH_INIT_BUCKETS (15)
#define HASH_RESIZE_MULTIPLIER (3)
#define HASH_INIT_RESIZE (90)

#define hash_KEY_LOC(e) ((void *)((hash_entry_t *)(e) + 1))

static          result_t
hash_str_overhead(const void *key,
		  unsigned *overhead)
{
  I(key);
  I(overhead);

  *overhead = str_MEM_NEEDED(str_LEN((const str_t *) key));
  return OK;
}

static result_t
hash_str_set_key(const void *key,
		 void *dest)
{
  I(key);
  I(dest);

  memcpy(dest, key, str_MEM_NEEDED(str_LEN((const str_t *) key)));

  return OK;
}

static result_t
hash_str_hash(const void *key,
	      hashvalue_t * hashvalue)
{
  I(key);
  I(hashvalue);

  *hashvalue = str_hashvalue((const str_t *) key);
  return OK;
}

static result_t
hash_str_equal(const void *a,
	       const void *b,
	       eq_result_t * equal)
{
  I(a);
  I(b);

  *equal = str_equal((const str_t *) a, (const str_t *) b);

  return OK;
}

const hash_procs_t hash_str_procs = {
  &hash_str_overhead,
  &hash_str_set_key,
  &hash_str_hash,
  &hash_str_equal
};

static result_t
hash_cstr_overhead(const void *key,
		   unsigned *overhead)
{
  I(key);
  I(overhead);

  *overhead = strlen((const char *) key) + 1;
  return OK;
}

static result_t
hash_cstr_set_key(const void *key,
		  void *dest)
{
  I(key);
  I(dest);
  strcpy(dest, key);
  return OK;
}

static result_t
hash_cstr_hash(const void *key,
	       hashvalue_t * hashvalue)
{
  const char     *k;

  hashvalue_t     hash;

  I(key);
  I(hashvalue);

  for (k = key, hash = 0; *k; ++k)
    hash += (hash << 3) + *k;

  if (!hash)
    hash = 1;
  *hashvalue = hash;
  return OK;
}

static result_t
hash_cstr_equal(const void *a,
		const void *b,
		eq_result_t * equal)
{
  I(a);
  I(b);

  if (strcmp((const char *) a, (const char *) b) == 0)
    *equal = EQ_IS_EQUAL;
  else
    *equal = EQ_NOT_EQUAL;

  return OK;
}

const hash_procs_t hash_cstr_procs = {
  &hash_cstr_overhead,
  &hash_cstr_set_key,
  &hash_cstr_hash,
  &hash_cstr_equal
};

static result_t
hash_ptr_overhead(const void *key,
		  unsigned *overhead)
{
  I(key);
  I(overhead);

  *overhead = sizeof(void *);
  return OK;
}

static result_t
hash_ptr_set_key(const void *key,
		 void *dest)
{
  I(key);
  I(dest);

  *(const void **) dest = key;

  return OK;
}

static result_t
hash_ptr_hash(const void *key,
	      hashvalue_t * hashvalue)
{
  I(key);
  I(hashvalue);

  *hashvalue = (unsigned) key >> HASH_PTR_SHIFT;
  return OK;
}

static result_t
hash_ptr_equal(const void *key,
	       const void *data,
	       eq_result_t * equal)
{
  I(key);
  I(data);

  if (key == *(const void *const *) data)
    *equal = EQ_IS_EQUAL;
  else
    *equal = EQ_NOT_EQUAL;

  return OK;
}

const hash_procs_t hash_ptr_procs = {
  &hash_ptr_overhead,
  &hash_ptr_set_key,
  &hash_ptr_hash,
  &hash_ptr_equal
};

static result_t
hash_uint32_overhead(const void *key,
		     unsigned *overhead)
{
  I(key);
  I(overhead);

  *overhead = sizeof(uint32_t);
  return OK;
}

static result_t
hash_uint32_set_key(const void *key,
		    void *dest)
{
  I(key);
  I(dest);

  *(uint32_t *) dest = (uint32_t) key;

  return OK;
}

static result_t
hash_uint32_hash(const void *key,
		 hashvalue_t * hashvalue)
{
  I(key);
  I(hashvalue);

  *hashvalue = (hashvalue_t) key;	/* just cast right across */
  return OK;
}

static result_t
hash_uint32_equal(const void *key,
		  const void *data,
		  eq_result_t * equal)
{
  I(key);
  I(data);

  if ((uint32_t) key == *(uint32_t *) data)
    *equal = EQ_IS_EQUAL;
  else
    *equal = EQ_NOT_EQUAL;

  return OK;
}

const hash_procs_t hash_uint32_procs = {
  &hash_uint32_overhead,
  &hash_uint32_set_key,
  &hash_uint32_hash,
  &hash_uint32_equal
};

result_t
hash_init(const hash_procs_t * procs,
	  hash_table_t * table)
{
  I(table);
  I(procs);

  I(procs->overhead);
  I(procs->set_key);
  I(procs->hash);
  I(procs->equal);

  table->buckets = NULL;
  table->n_buckets = 0;
  table->n_entries = 0;
  table->next_resize = 0;
  table->procs = procs;

  return OK;
}

result_t
hash_destroy(hash_table_t * table)
{
  unsigned        bucket;

  hash_entry_t   *entry,
                 *next;

  I(table);
  I(!table->n_buckets || table->buckets);

  for (bucket = 0; bucket < table->n_buckets; bucket++)
    for (entry = table->buckets[bucket]; entry; entry = next) {
      next = entry->next;
      free(entry);
    }

  if (table->buckets)
    free(table->buckets);

  return OK;
}


static result_t
hash_resize(hash_table_t * table)
{
  hash_entry_t  **bucket,
                 *current_entry,
                 *next,
                **new_buckets;

  unsigned        new_n_buckets,
                  bucket_number;

  I(table);

  if (table->n_buckets)
    new_n_buckets = table->n_buckets * HASH_BUCKET_MULTIPLIER;
  else
    new_n_buckets = HASH_INIT_BUCKETS;

  new_buckets = (hash_entry_t **) calloc(new_n_buckets,
					 sizeof(hash_entry_t *));
  if (!new_buckets)
    return errno;

  if (table->n_buckets && table->buckets)
    for (bucket_number = 0;
	 bucket_number < table->n_buckets;
	 bucket_number++)
      for (current_entry = table->buckets[bucket_number];
	   current_entry;
	   current_entry = next) {
	bucket = new_buckets
	  + (current_entry->hashvalue % new_n_buckets);

	next = current_entry->next;

	current_entry->bucket = bucket;
	current_entry->next = *bucket;
	*bucket = current_entry;
      }
  free(table->buckets);
  table->n_buckets = new_n_buckets;
  table->buckets = new_buckets;
  if (table->next_resize)
    table->next_resize *= HASH_RESIZE_MULTIPLIER;
  else
    table->next_resize = HASH_INIT_RESIZE;

  return OK;
}

result_t
hash_find(hash_table_t * table,
	  const void *key,
	  hash_insert_t insert,
	  hash_did_insert_t * inserted,
	  hash_entry_t ** entry_ret)
{
  result_t        result;

  hashvalue_t     hashvalue;

  eq_result_t     key_equal;

  hash_entry_t   *entry;

  I(table);

  if (entry_ret)
    *entry_ret = NULL;

  result = table->procs->hash(key, &hashvalue);
  if (result)
    return result;

  entry = NULL;

  /* Find the entry chain... */
  if (table->n_buckets && table->buckets)
    entry = table->buckets[hashvalue % table->n_buckets];

  /* Walk down to the entry... */
  while (entry
	 && !(result = table->procs->equal(key,
					   hash_KEY_LOC(entry),
					   &key_equal))
	 && key_equal == EQ_NOT_EQUAL)
    entry = entry->next;

  if (result)
    return result;

  /* If we don't have the entry, try to insert it */
  if (!entry) {
    if (insert == HASH_NO_INSERT) {
      /* We don't have it and we can't insert it.  Bummer. */
      return HASH_NOT_FOUND;
    } else {
      /* Insert a new entry for this key */
      unsigned        overhead;

      hash_entry_t  **bucket;

      result = table->procs->overhead(key, &overhead);
      if (result)
	return result;

      if (table->next_resize <= table->n_entries) {
	result = hash_resize(table);
	if (result)
	  return result;
      }
      entry = malloc(sizeof(hash_entry_t) + overhead);
      if (!entry)
	return errno;

      result = table->procs->set_key(key, hash_KEY_LOC(entry));
      if (result) {
	free(entry);
	return result;
      }
      bucket = table->buckets + (hashvalue % table->n_buckets);
      entry->hashvalue = hashvalue;
      entry->table = table;
      entry->data = NULL;
      entry->bucket = bucket;
      entry->next = *bucket;
      *bucket = entry;

      if (inserted)
	*inserted = HASH_DID_INSERT;

      table->n_entries++;
    }
  } else {
    /* We didn't have to insert it */
    if (inserted)
      *inserted = HASH_DID_NOT_INSERT;
  }

  /* We shouldn't be able to get here unless entry's set */
  I(entry);

  if (entry_ret)
    *entry_ret = entry;

  return OK;
}

result_t
hash_delete_entry(hash_entry_t * entry)
{
  I(entry);
  I(entry->table);
  I(entry->bucket);

  if (*(entry->bucket) == entry)
    *(entry->bucket) = entry->next;
  else {
    hash_entry_t   *ent = *(entry->bucket);

    while (ent->next && ent->next != entry)
      ent = ent->next;
    I(ent);
    ent->next = entry->next;
  }

  entry->table->n_entries--;

  free(entry);

  return OK;
}

result_t
hash_iter_begin(hash_table_t * table,
		hash_iter_t * iter)
{
  I(table);
  I(iter);

  iter->table = table;
  iter->next_bucket = 0;
  iter->next_entry = NULL;

  return OK;
}

result_t
hash_iter_next(hash_iter_t * iter,
	       hash_entry_t ** entry)
{
  I(iter);
  I(entry);

  while (!iter->next_entry) {
    if (iter->table->n_buckets <= iter->next_bucket)
      return HASH_ITER_END;
    iter->next_entry = iter->table->buckets[iter->next_bucket++];
  }

  *entry = iter->next_entry;
  iter->next_entry = iter->next_entry->next;
  return OK;
}
