/* rselock.c - read+shared+exclusive lock implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define RSELOCK_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include "rselock.h"
#include "util_err.h"

typedef struct {
  pthread_mutex_t *mutex;
  pthread_cond_t *cond;
  pthread_t       thread_id;
}               writer_queue_elem_t;

result_t
rselock_init(rselock_t * lock)
{
  result_t        result = OK;

  I(lock);

  result = pthread_mutex_init(&lock->mutex, NULL);
  if (result)
    goto cleanup;

  lock->n_readers = 0;
  lock->allow_more_readers = TRUE;
  lock->is_write_locked = FALSE;

  result = pthread_cond_init(&lock->reader_cond, NULL);
  if (result)
    goto cleanup_mutex;

  result = queue_init(&lock->writer_queue);
  if (result)
    goto cleanup_reader_cond;

  return result;

cleanup_reader_cond:
  pthread_cond_destroy(&lock->reader_cond);

cleanup_mutex:
  pthread_mutex_destroy(&lock->mutex);

cleanup:
  return result;
}

result_t
rselock_destroy(rselock_t * lock)
{
  I(lock);

  if (I_DEFAULT_GUARD) {
    pthread_mutex_lock(&lock->mutex);
    I(!lock->n_readers);
    I(!lock->is_write_locked);
    I(lock->allow_more_readers);
    I(queue_ISEMPTY(&lock->writer_queue));
    pthread_mutex_unlock(&lock->mutex);
  }
  pthread_cond_destroy(&lock->reader_cond);
  pthread_mutex_destroy(&lock->mutex);

  return OK;
}

result_t
rselock_read_lock(rselock_t * lock)
{
  I(lock);

  pthread_mutex_lock(&lock->mutex);

  while (!lock->allow_more_readers)
    pthread_cond_wait(&lock->reader_cond, &lock->mutex);

  I(lock->allow_more_readers);

  lock->n_readers++;

  pthread_mutex_unlock(&lock->mutex);

  return OK;
}

result_t
rselock_read_unlock(rselock_t * lock)
{
  I(lock);

  pthread_mutex_lock(&lock->mutex);

  I(lock->n_readers);

  lock->n_readers--;

  if (!lock->n_readers && !lock->allow_more_readers) {
    /*
     * We're the last reader to be leaving this lock; We want to wake up
     * whoever's waiting for it...
     */
    I(lock->boost_cond);
    pthread_cond_signal(lock->boost_cond);
  }
  pthread_mutex_unlock(&lock->mutex);

  return OK;
}

result_t
rselock_shared_lock_signal(rselock_t * lock,
			   pthread_mutex_t * mutex,
			   pthread_cond_t * cond)
{
  result_t        result = OK;

  I(lock);
  I(mutex);
  I(cond);

  pthread_mutex_lock(&lock->mutex);

  if (!lock->is_write_locked) {
    lock->is_write_locked = TRUE;
    lock->current_write_thread = pthread_self();
  } else {
    writer_queue_elem_t *qelem;

    qelem = (writer_queue_elem_t *) malloc(sizeof(writer_queue_elem_t));
    if (!qelem) {
      result = errno;
      goto cleanup_lock;
    }
    qelem->mutex = mutex;
    qelem->cond = cond;
    qelem->thread_id = pthread_self();

    result = queue_push(&lock->writer_queue, qelem);
    if (result) {
      free(qelem);
      goto cleanup_lock;
    }
    result = RSELOCK_WOULDBLOCK;/* caller needs to wait for the lock */
  }

cleanup_lock:
  pthread_mutex_unlock(&lock->mutex);

  return result;
}

result_t
rselock_shared_lock(rselock_t * lock)
{
  result_t        result = OK;

  pthread_mutex_t mutex;

  pthread_cond_t  cond;

  I(lock);

  result = pthread_mutex_init(&mutex, NULL);
  if (result)
    goto cleanup;

  result = pthread_cond_init(&cond, NULL);
  if (result)
    goto cleanup_mutex;

  pthread_mutex_lock(&mutex);

  result = rselock_shared_lock_signal(lock, &mutex, &cond);
  if (result && result != RSELOCK_WOULDBLOCK)
    goto cleanup_lock;

  if (result == RSELOCK_WOULDBLOCK)
    result = pthread_cond_wait(&cond, &mutex);

  if (I_DEFAULT_GUARD) {
    pthread_mutex_lock(&lock->mutex);
    I(lock->is_write_locked);
    I(lock->current_write_thread == pthread_self());
    pthread_mutex_unlock(&lock->mutex);
  }
cleanup_lock:
  pthread_mutex_unlock(&mutex);

  pthread_cond_destroy(&cond);

cleanup_mutex:
  pthread_mutex_destroy(&mutex);

cleanup:
  return result;
}

/* Utility function for rselock_shared_unlock and rselock_exclusive_unlock.
 * Assumed lock->mutex is held. */
static result_t
rselock_protected_unlock(rselock_t * lock)
{
  result_t        result = OK;

  writer_queue_elem_t *qelem = NULL;

  I(lock);

  I(lock->is_write_locked);
  I(pthread_equal(lock->current_write_thread, pthread_self()));

  lock->is_write_locked = FALSE;
  lock->allow_more_readers = TRUE;

  result = queue_pop(&lock->writer_queue, (void **) &qelem);
  if (result == QUEUE_EMPTY) {
    result = OK;
  } else if (result == OK) {
    /*
     * We need to wake up someone who was patiently waiting for this lock.
     */
    I(qelem);
    I(qelem->mutex);
    I(qelem->cond);

    pthread_mutex_lock(qelem->mutex);	/* make sure he's waiting... */
    pthread_cond_signal(qelem->cond);	/* wake him up... */
    lock->current_write_thread = qelem->thread_id;	/* give him the lock, */
    lock->is_write_locked = TRUE;
    pthread_mutex_unlock(qelem->mutex);	/* and let him run. */
    free(qelem);
  }
  return result;
}

result_t
rselock_shared_unlock(rselock_t * lock)
{
  result_t        result;

  I(lock);

  pthread_mutex_lock(&lock->mutex);

  I(lock->allow_more_readers);

  result = rselock_protected_unlock(lock);

  pthread_mutex_unlock(&lock->mutex);

  return result;
}

result_t
rselock_exclusive_lock(rselock_t * lock)
{
  result_t        result;

  I(lock);

  result = rselock_shared_lock(lock);
  if (result)
    return result;
  result = rselock_boost(lock);
  if (result)
    rselock_shared_unlock(lock);
  return result;
}

result_t
rselock_exclusive_unlock(rselock_t * lock)
{
  result_t        result;

  I(lock);

  pthread_mutex_lock(&lock->mutex);

  I(!lock->allow_more_readers);

  /* Wake up those readers... */
  pthread_cond_broadcast(&lock->reader_cond);

  /* And actually do the unlock */
  result = rselock_protected_unlock(lock);

  pthread_mutex_unlock(&lock->mutex);

  return result;
}

result_t
rselock_boost(rselock_t * lock)
{
  result_t        result = OK;

  pthread_cond_t  cond;

  I(lock);

  pthread_mutex_lock(&lock->mutex);

  I(lock->allow_more_readers);
  I(lock->is_write_locked);
  I(pthread_equal(lock->current_write_thread, pthread_self()));

  if (lock->n_readers) {
    /* Wait for them to complete... */
    result = pthread_cond_init(&cond, NULL);
    if (result)
      goto cleanup;
    lock->allow_more_readers = FALSE;
    lock->boost_cond = &cond;
    pthread_cond_wait(&cond, &lock->mutex);
  } else {
    lock->allow_more_readers = FALSE;
  }

  I(!lock->n_readers);

cleanup:
  pthread_mutex_unlock(&lock->mutex);
  return result;
}

result_t
rselock_unboost(rselock_t * lock)
{
  I(lock);

  pthread_mutex_lock(&lock->mutex);

  I(!lock->allow_more_readers);

  lock->allow_more_readers = TRUE;
  pthread_cond_broadcast(&lock->reader_cond);

  pthread_mutex_unlock(&lock->mutex);

  return OK;
}

rselock_locked_t
rselock_check_lock(rselock_t * lock)
{
  rselock_locked_t status;

  I(lock);

  pthread_mutex_lock(&lock->mutex);

  if (!lock->is_write_locked) {
    if (!lock->n_readers)
      status = RSELOCK_UNLOCKED;
    else
      status = RSELOCK_READ_LOCKED;
  } else if (!pthread_equal(lock->current_write_thread,
			    pthread_self())) {
    if (lock->allow_more_readers)
      status = RSELOCK_SHARED_LOCKED;
    else
      status = RSELOCK_EXCLUSIVE_LOCKED;
  } else if (lock->allow_more_readers)
    status = RSELOCK_HAVE_SHARED_LOCK;
  else
    status = RSELOCK_HAVE_EXCLUSIVE_LOCK;

  pthread_mutex_unlock(&lock->mutex);

  return status;
}
