/* rselock.h - read+shared+exclusive lock interface
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef RSELOCK_H
#define RSELOCK_H

/* An RSE lock is a lock which can be read locked, shared locked, or
 * exclusively locked.  A read lock is compatible with read locks and
 * shared locks; a shared lock is compatible with read locks, and an
 * exclusive lock isn't compatible with anything.  A shared lock
 * may be boosted to an exclusive lock (blocking new readers and
 * waiting for current readers to release their lock).
 *
 * One oddity: it's easy for a thread to wait for multiple locks
 * (either for a shared lock, or when boosting to an exclusive lock).
 * It needs to supply its own pthread_mutex_t and pthread_cond_t, but
 * this way, a thread can wait for its locks in parallel, instead of
 * serially.  This turns out to be a major win for some tasks...
 *
 * We don't implement priority inheritance for now; we may want to,
 * at some point in the future.  */

#include <pthread.h>
#include "queue.h"

typedef struct {
  pthread_mutex_t mutex;
  unsigned        n_readers;
  unsigned        allow_more_readers:1;
  unsigned        is_write_locked:1;
  pthread_cond_t  reader_cond;
  q_t             writer_queue;
  pthread_t       current_write_thread;
  pthread_cond_t *boost_cond;
}               rselock_t;

/* Creation/destruction */
extern          result_t rselock_init(rselock_t * lock);

extern result_t rselock_destroy(rselock_t * lock);

/* Read lock */
extern result_t rselock_read_lock(rselock_t * lock);

extern result_t rselock_read_unlock(rselock_t * lock);

/* Shared lock */
extern result_t rselock_shared_lock(rselock_t * lock);

/* shared_lock_signal returns RSELOCK_WOULDBLOCK if
 * it wasn't able to get the lock immediately; the caller should
 * call pthread_cond_wait on the mutex and the cond. */
extern result_t
rselock_shared_lock_signal(rselock_t * lock,
			   pthread_mutex_t * mutex,
			   pthread_cond_t * cond);

extern result_t rselock_shared_unlock(rselock_t * lock);

/* Exclusive lock */
extern result_t rselock_exclusive_lock(rselock_t * lock);

extern result_t rselock_exclusive_unlock(rselock_t * lock);

/* Boosting */
extern result_t rselock_boost(rselock_t * lock);

extern result_t rselock_unboost(rselock_t * lock);

extern result_t
rselock_boost_signal(rselock_t * lock,
		     pthread_mutex_t * mutex,
		     pthread_cond_t * cond);

/* Checking */
typedef enum {
  RSELOCK_UNLOCKED,		/* nobody has this locked */
  RSELOCK_READ_LOCKED,		/* there are outstanding readlocks */
  RSELOCK_SHARED_LOCKED,	/* locked by someone else */
  RSELOCK_EXCLUSIVE_LOCKED,	/* locked by someone else */
  RSELOCK_HAVE_SHARED_LOCK,	/* thread owns shared lock */
  RSELOCK_HAVE_EXCLUSIVE_LOCK	/* thread owns exclusive lock */
}               rselock_locked_t;

/* Note that if this returns RSELOCK_UNLOCKED, it's pretty much
 * useless unless you know that nothing else could possibly acquire
 * the lock.  If this returns RSELOCK_READ_LOCKED,
 * RSELOCK_SHARED_LOCKED, or RSELOCK_EXCLUSIVE_LOCKED, it's useless
 * unless you know the locker can't possibly release its lock.
 * Note that if RSELOCK_SHARED_LOCKED or RSELOCK_HAVE_SHARED_LOCK
 * is returned, there may be other threads with reader locks. */
extern          rselock_locked_t
                rselock_check_lock(rselock_t * lock);

#endif				/* RSELOCK_H */
