/* sllist.h - singly-linked-list interface
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#ifndef SLLIST_H
#define SLLIST_H

typedef struct sllist_s {
  struct sllist_s *next;
  void           *data;
}               sllist_t;

#define sllist_NEXT(s) ((s)->next)
#define sllist_DATA(s) ((s)->data)

#define sllist_SET_NEXT(s, v) ((s)->next = (v))
#define sllist_SET_DATA(s, d) ((s)->data = (d))

/* Allocate an sllist_t */
extern          result_t
sllist_alloc(void *data,
	     sllist_t * next,
	     sllist_t ** result);

/* Apply a function to all the data in the sllist_t.  The apply stops
 * and returns the result of the apply proc if the apply proc returns
 * anything other than OK or SLLIST_APPLY_STOP; if the apply proc
 * returns SLLIST_APPLY_STOP, the apply will stop, and return OK.  The
 * apply also stops, returning SLLIST_LOOP_DETECTED, if a loop is
 * detected in the list; note the the function may be applied more
 * than once to a given piece of data in this case. */

typedef result_t sllist_apply_proc_t(sllist_t * data, void *rock);

extern result_t
sllist_apply(sllist_t * list,
	     sllist_apply_proc_t * proc,
	     void *rock);

/* Free one sllist_t node */
extern result_t sllist_free_node(sllist_t ** list);

/* Free an entire list */
extern result_t sllist_free(sllist_t ** list);

/* Perform a simple list reversal */
extern result_t sllist_reverse(sllist_t ** list);

/* sllist_compare_proc_t is used for operations on ordered lists; it
 * allows the caller to define an arbitrary ordering on the elements
 * in what's hoped to be a reasonably flexible way.  If a
 * sllist_compare_proc_t returns non-OK, the result will bubble up;
 * elements may be shuffled, but no elements will be lost. */
typedef result_t
sllist_compare_proc_t(const sllist_t * a,
		      const sllist_t * b,
		      void *rock,
		      compare_result_t * compare_result);

/* Merges together two lists, assuming that they're sorted.  If the
 * compare_proc returns non-OK, the lists will be returned appended
 * together in some random order; they'll still be destroyed. */
extern result_t
sllist_merge(sllist_t * a, sllist_t * b,
	     sllist_compare_proc_t * comparator,
	     void *rock,
	     sllist_t ** dest);

/* Sorts a list */
extern result_t
sllist_sort(sllist_t ** list,
	    sllist_compare_proc_t * comparator,
	    void *rock);

typedef sllist_t *sllist_iterator_t;

#define sllist_ITER_BEGIN(s, i) ((*i) = (s))
#define sllist_ITER_ATEND(i) ((*i) == NULL)
#define sllist_ITER_INCR(i) ((*i) = (*i)->next)
#define sllist_ITER_VALUE(i) (*i)
#endif				/* SLLIST_H */
