/* str.c - length-string implementation
 * Rob Earhart
 */
/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Carnegie Mellon
University not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR
ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
******************************************************************/

#define STR_C
#include <config.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include "str.h"
#include "util_err.h"

result_t
str_alloc(uint32_t length,
	  const char *buf,
	  str_t ** str)
{
  I(str);

  if (!*str) {
    /* Allocate it... */
    *str = (str_t *) malloc(str_MEM_NEEDED(length));
    if (!*str)
      return errno;
  }
  /* And fill it in */
  str_SET_LEN(*str, length);
  if (buf)
    memcpy(str_DATA(*str), buf, length);

  /* Ensure a terminator, and clear the rest of the str */
  I(length < str_MEM_NEEDED(length) - sizeof(str_t));
  memset(str_DATA(*str) + length, 0,
	 str_MEM_NEEDED(length) - sizeof(str_t) - length);

  I(*str);

  return OK;
}

result_t
str_copy(const str_t * in, str_t ** out)
{
  I(in);
  I(out);

  return str_alloc(str_LEN(in), str_DATA(in), out);
}

result_t
str_copy_cstr(const char *in, str_t ** out)
{
  I(in);
  I(out);

  return str_alloc(strlen(in), in, out);
}

result_t
str_append(str_t ** dest,...)
{
  va_list         ap;

  uint32_t        len;

  str_append_str_type_t t;

  char           *pos;

  const str_t    *str;

  const char     *s;

  char            c;

  I(dest);

  len = 0;
  va_start(ap, dest);
  do {
    t = va_arg(ap, str_append_str_type_t);
    switch (t) {
    case STR_APPEND_DONE:
      break;
    case STR_APPEND_STR_T:
      str = va_arg(ap, str_t *);
      len += str_LEN(str);
      break;
    case STR_APPEND_C_STRING:
      s = va_arg(ap, const char *);
      len += strlen(s);
      break;
    case STR_APPEND_CHAR:
      c = va_arg(ap, char);
      len++;
      break;
    default:
      return STR_APPEND_BAD_STR_TYPE;
    }
  } while (t != STR_APPEND_DONE);

  va_end(ap);

  *dest = (str_t *) malloc(str_MEM_NEEDED(len));
  if (!*dest)
    return errno;

  str_SET_LEN(*dest, len);
  pos = str_DATA(*dest);

  va_start(ap, dest);
  do {
    t = va_arg(ap, str_append_str_type_t);
    switch (t) {
    case STR_APPEND_DONE:
      break;
    case STR_APPEND_STR_T:
      str = va_arg(ap, str_t *);
      len = str_LEN(str);
      memcpy(pos, str_DATA(str), len);
      pos += len;
      break;
    case STR_APPEND_C_STRING:
      s = va_arg(ap, const char *);
      len = strlen(s);
      memcpy(pos, s, len);
      pos += len;
      break;
    case STR_APPEND_CHAR:
      /* This little hack is necessary because gcc (under Solaris, at
       * least) does the wrong thing if you use va_arg(ap, char); it
       * grabs the wrong byte off the stack.  If you grab an unsigned
       * and cast, though, it's fine.  Go figure. */
      c = (char) va_arg(ap, unsigned);
      *pos = c;
      pos++;
      break;
    default:
      free(*dest);
      *dest = NULL;
      return STR_APPEND_BAD_STR_TYPE;
    }
  } while (t != STR_APPEND_DONE);

  *pos++ = '\0';

  va_end(ap);

  return OK;
}

result_t
str_join(const str_t * prefix,
	 const str_t * seperator,
	 const str_t * suffix,
	 const str_t * const * begin,
	 const str_t * const * end,
	 str_t ** dest)
{
  const str_t    *const * it;

  uint32_t        len;

  char           *pos;

  /* Calculate the amount of space required... */
  len = 0;
  if (prefix)
    len += str_LEN(prefix);
  for (it = begin; it < end; ++it) {
    I(*it);
    if (it != begin && seperator)
      len += str_LEN(seperator);
    len += str_LEN(*it);
  }
  if (suffix)
    len += str_LEN(suffix);

  /* Allocate the storage... */
  *dest = (str_t *) malloc(str_MEM_NEEDED(len));
  if (!*dest)
    return errno;

  /* And fill it in. */
  str_SET_LEN(*dest, len);
  pos = str_DATA(*dest);

  if (prefix) {
    memcpy(pos, str_DATA(prefix), str_LEN(prefix));
    pos += str_LEN(prefix);
  }
  for (it = begin; it < end; ++it) {
    I(*it);
    if (it != begin && seperator) {
      memcpy(pos, str_DATA(seperator), str_LEN(seperator));
      pos += str_LEN(seperator);
    }
    memcpy(pos, str_DATA(*it), str_LEN(*it));
    pos += str_LEN(*it);
  }
  if (suffix) {
    memcpy(pos, str_DATA(suffix), str_LEN(suffix));
    pos += str_LEN(suffix);
  }
  /* And cap it off, for sanity */
  *pos = '\0';

  return OK;
}

result_t
str_free(str_t ** str)
{
  I(str);

  if (*str) {
    free(*str);
    *str = NULL;
  }
  return OK;
}

eq_result_t
str_equal(const str_t * a, const str_t * b)
{
  I(a);
  I(b);

  if (a == b
      || (a->len == b->len
	  && !memcmp(str_DATA(a), str_DATA(b), str_LEN(a))))
    return EQ_IS_EQUAL;
  else
    return EQ_NOT_EQUAL;
}

eq_result_t
str_equal_cstr(const str_t * str, const char *cstr)
{
  unsigned        len;

  I(str);
  I(cstr);

  len = strlen(cstr);
  if (len == str_LEN(str)
      && !memcmp(str_DATA(str), cstr, len))
    return EQ_IS_EQUAL;
  else
    return EQ_NOT_EQUAL;
}

hashvalue_t
str_hashvalue(const str_t * str)
{
  hashvalue_t     hashvalue = 0;

  const char     *lupe,
                 *end;

  I(str);

  lupe = str_DATA(str);
  end = lupe + str_LEN(str);
  while (lupe < end) {
    hashvalue += (hashvalue << 3) + *lupe;
    lupe++;
  }

  if (!hashvalue)
    hashvalue = 1;

  return hashvalue;
}

compare_result_t
str_compare(const str_t * a,
	    const str_t * b)
{
  int             comp;

  I(a);
  I(b);

  if (a == b)
    return COMPARE_EQUAL;

  comp = memcmp(str_DATA(a), str_DATA(b),
		str_LEN(a) < str_LEN(b) ? str_LEN(a) : str_LEN(b));
  if (comp < 0)
    return COMPARE_LESS;
  if (0 < comp)
    return COMPARE_GREATER;

  comp = str_LEN(a) - str_LEN(b);
  if (comp < 0)
    return COMPARE_LESS;
  if (0 < comp)
    return COMPARE_GREATER;

  return COMPARE_EQUAL;
}


str_is_utf8_t
str_is_UTF8(const str_t * str)
{
  const char     *c,
                 *end;

  for (c = str_DATA(str), end = c + str_LEN(str);
       c < end;
       ++c) {
    unsigned        seqlen = 0;

    if (!*c)
      return STR_NOT_UTF8;	/* no NULLs in UTF8 strings */

    /* how many octets? */
    /*
     * This could probably be faster.  TODO: benchmark this code; try
     * unrolling the loop, and try a lookup table, comparing ints (better
     * alignment) shorts, and bytes (better packing / fewer cache line hits).
     */
    while (*c & (0x80 >> seqlen))
      ++seqlen;
    if (seqlen == 0)
      continue;			/* valid US-ASCII character */
    if (seqlen == 1)
      return STR_NOT_UTF8;	/* illegal in the len octet */
    if (seqlen > 6)
      return STR_NOT_UTF8;	/* illegal */

    while (--seqlen) {
      ++c;
      if (end <= c)
	return STR_NOT_UTF8;	/* over the end */
      if ((*c & 0xC0) != 0x80)
	return STR_NOT_UTF8;	/* needed a 10 octet */
    }
  }
  return STR_IS_UTF8;
}
